# Copyright (C) Martin Singer <martin.singer@web.de>
#
# This file is part of PDF Chain.
#
# PDF Chain is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PDF Chain is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PDF Chain.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Martin Singer


'''
@brief Terminal Color Commands.
@details
    This module contains constant strings with the escape sequences
    for various colors and styles in the terminal.

@see <https://en.wikipedia.org/wiki/ANSI_escape_code#3/4_bit>
'''

__ESC       = '\033'

RESET     = __ESC + '[0m'
#BOLD      = __ESC + '[1m'
#FAINT     = __ESC + '[2m'
#ITALIC    = __ESC + '[3m'
#UNDERLINE = __ESC + '[4m'

#FG_BLACK          = __ESC + '[30m'
#FG_RED            = __ESC + '[31m'
#FG_GREEN          = __ESC + '[32m'
#FG_YELLOW         = __ESC + '[33m'
#FG_BLUE           = __ESC + '[34m'
#FG_MAGENTA        = __ESC + '[35m'
#FG_CYAN           = __ESC + '[36m'
#FG_WHITE          = __ESC + '[37m'
#FG_BRIGHT_BLACK   = __ESC + '[90m'
FG_BRIGHT_RED     = __ESC + '[91m'
FG_BRIGHT_GREEN   = __ESC + '[92m'
FG_BRIGHT_YELLOW  = __ESC + '[93m'
FG_BRIGHT_BLUE    = __ESC + '[94m'
FG_BRIGHT_MAGENTA = __ESC + '[95m'
FG_BRIGHT_CYAN    = __ESC + '[96m'
#FG_BRIGHT_WHITE   = __ESC + '[97m'

#BG_BLACK          = __ESC + '[40m'
#BG_RED            = __ESC + '[41m'
#BG_GREEN          = __ESC + '[42m'
#BG_YELLOW         = __ESC + '[43m'
#BG_BLUE           = __ESC + '[44m'
#BG_MAGENTA        = __ESC + '[45m'
#BG_CYAN           = __ESC + '[46m'
#BG_WHITE          = __ESC + '[47m'
#BG_BRIGHT_BLACK   = __ESC + '[100m'
#BG_BRIGHT_RED     = __ESC + '[101m'
#BG_BRIGHT_GREEN   = __ESC + '[102m'
#BG_BRIGHT_YELLOW  = __ESC + '[103m'
#BG_BRIGHT_BLUE    = __ESC + '[104m'
#BG_BRIGHT_MAGENTA = __ESC + '[105m'
#BG_BRIGHT_CYAN    = __ESC + '[106m'
#BG_BRIGHT_WHITE   = __ESC + '[107m'




#!/usr/bin/env python3

# Copyright (C) Martin Singer <martin.singer@web.de>
#
# This file is part of PDF Chain.
#
# PDF Chain is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PDF Chain is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PDF Chain.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Martin Singer

'''
@brief Project File Path Module.
@details
    This module manages the project file paths
    and contains defined filenames.
'''

from os.path import abspath, dirname, exists, join, sep, split
import xdg.BaseDirectory


def __proof_path_exists(path):
    """
    """
    if exists(path):
        return path
    raise Exception("The path '{}' was not found!".format(path))


def __get_base_dir():
    """
    @brief Finds and returns the data base directory path (`share`).
    @details
        This function searches the data directory belonging to the executed program.
        This works for installations where the data directory is called "share"
        and shares a common directory tree with the execution directory.
    @par Explaining the Algorithm
        This function determines the directory in which it is executed
        and scans the directory tree backwards in a loop step by step.
        At each step it checks if it contains the data directory "share".
        If "share" is found, it is assumed to contain the data directory
        and the path of the data directory is returned.
        If the loop has arrived at the root directory (`/`),
        no data directory was found and the loop is aborted
        and an exception is thrown.
        This simple algorithm should work for different installation paths.
    @code List Data Directories
        import xdg.BaseDirectory
        print(xdg.BaseDirectory.xdg_data_dirs)
    @par Known Installation Cases
        - Development
            - `~/Projects/*/pdfchain/src`
            - `~/Projects/*/pdfchain/share`
        - PIP User Installation
            - `~/.local/lib/python<version>/site-packages`
            - `~/.local/share`
        - Flatpak User Installation
            - `~/.local/share/flatpak/*/exports/bin
            - `~/.local/share/flatpak/*/exports/share`
        - Flatpak System Installation
            - `/var/lib/flatpak/*/exports/bin`
            - `/var/lib/flatpak/*/exports/share`
        - Local System Installation (`sudo make install`)
            - `/usr/local/bin`
            - `/usr/local/share`
        - System Installation (package manager)
            - `/usr/bin`
            - `/usr/share`
    """
    head_path = (split(abspath(dirname(__file__))))[0]
    while head_path != sep:
        base_path = join(head_path, "share")
        if exists(base_path):
            print("BasePath: '{}'".format(base_path))
            return base_path
        head_path = (split(head_path))[0]
    raise Exception("No base directory found!")


def __get_data_dir(base_dir_path, data_dir_name):
    """
    @brief Returns the requested data directory.
    """
    data_dir_path = abspath(join(base_dir_path, data_dir_name))
    if exists(data_dir_path):
        return data_dir_path
    raise Exception("No data directory '{}' in '{}' found!".format(
        data_dir_name, base_dir_path))


def get_locale_dir():
    """
    @brief Returns locale directory path.
    """
    return __LOCALE_DIR


def get_ui_file(filename):
    return __proof_path_exists(join(__UI_DIR, filename))


#def get_icon_file(filename):
#    return __proof_path_exists(join(__ICONS_DIR, filename))


#def get_pixmap_file(filename):
#    return __proof_path_exists(join(__PIXMAPS_DIR, filename))


def get_config_file():
    """
    @brief Returns the path of the users configuration file.
    """
    return __CONFIG_FILE


__BASE_DIR    = __get_base_dir()
__UI_DIR      = __get_data_dir(__BASE_DIR, join("pdfchain", "ui"))
# __LOCALE_DIR  = __get_data_dir(__BASE_DIR, "locale")
__ICONS_DIR   = __get_data_dir(__BASE_DIR, "icons")
__PIXMAPS_DIR = __get_data_dir(__BASE_DIR, "pixmaps")

__CONFIG_FILE = __proof_path_exists(join(xdg.BaseDirectory.xdg_config_home, "pdfchain.ini"))



# Copyright (C) Martin Singer <martin.singer@web.de>
#
# This file is part of PDF Chain.
#
# PDF Chain is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PDF Chain is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PDF Chain.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Martin Singer


import os
import re # RegEx parser

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

#import pdfchain.strings as strings
import pdfchain.cmd_colors as color


'''
 @brief   Command Line Interface
 @details
    This module contains functions for outputs on the Command Line Interface (CLI).
'''


STR_NAME = "pdfchain"


def init(str_name):
    '''
    @brief Initializes this module.
    @details
        - sets the global program name variable, used for the CLI messages

    @param[in] str_name The program name string
    '''
    global STR_NAME
    STR_NAME = str_name


def debug_message(str_message):
    '''
    @brief Prints debug messages to the command line.
    @details
        Only calls the `cli.message` function with the `Gtk.MessageType.OTHER`,
        which will be interpreted and printed as "DEBUG" message.

    @param[in] str_message The message string
    '''
    message(Gtk.MessageType.OTHER, str_message)


def message(message_type, str_message):
    '''
    @brief Prints colored user messages to the command line.
    @details
        This function is called when a Gtk.InfoBar message appears.
        It prints the message with the leading program name
        and a colored message type indicator to the command line.

    @par Message Types
        - `Gtk.MessageType.INFO`     --> "INFO"
        - `Gtk.MessageType.WARNING`  --> "WARNING"
        - `Gtk.MessageType.QUESTION` --> "QUESTION"
        - `Gtk.MessageType.ERROR`    --> "ERROR"
        - `Gtk.MessageType.OTHER`    --> "DEBUG"

    @see window.AppWindowHandler._infobar_message()

    @param[in] message_type The `Gtk.MessageType`
    @param[in] str_message  The message string
    '''
    if message_type is Gtk.MessageType.INFO:
        str_type = color.FG_BRIGHT_BLUE + r"INFO" + color.RESET
    elif message_type is Gtk.MessageType.WARNING:
        str_type = color.FG_BRIGHT_YELLOW + r"WARNING" + color.RESET
    elif message_type is Gtk.MessageType.QUESTION:
        str_type = color.FG_BRIGHT_GREEN + r"QUESTION" + color.RESET
    elif message_type is Gtk.MessageType.ERROR:
        str_type = color.FG_BRIGHT_RED + r"ERROR" + color.RESET
    elif message_type is Gtk.MessageType.OTHER:
        str_type = color.FG_BRIGHT_MAGENTA + r"DEBUG" + color.RESET
    else:
        pass # write unknown messages without type indicator to the command line

    if str_type:
        str_type += ": "

    str_output = STR_NAME + ": " + str_type + str_message
    print(str_output)


def help():
    '''
    @brief Print program usage informations.
    @details
        Print informations about the program options to standard output.

    @TODO Command line options are not implemented yet!
    @TODO Get some strings from the UI (app name, ...) and build system (version, ...)
    '''
    print("""\
        NAME
            pdfchain - a graphical user interface for the PDF Toolkit

        SYNOPSIS
            pdfchain [OPTION]... [FILE]...

        DESCRIPTION
            Manages PDF documents easy for processing with the `pdftk` command line program.

            Mandatory arguments to long options are mandatory for short options too.

            -p --path=PATH    Path to document file or directory
    """)


def execute(str_command):
    '''
    @brief Execute the PDFtk command.
    @details
        Execute the command with `os.system()` function.

    @TODO Run the command execution in a separate Vte.Terminal() widget thread

    @return     The PDFtk error code
    @retval   0 Everything worked fine
    @retval 256 sh: pdftk: command not found
    @retval 768 pdftk: input PDF is not an acroform, so its fields were not filled
    '''
    print("{}".format(str_command))
    exit_code = os.system(str_command)
    return exit_code



# Copyright (C) Martin Singer <martin.singer@web.de>
#
# This file is part of PDF Chain.
#
# PDF Chain is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PDF Chain is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PDF Chain.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Martin Singer

"""
@brief The Application Module.
@details
    The application manages the various windows and dialogs.
    It contains the preferences and about dialog
    as well as multiple application window instances.
"""

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import GLib
from gi.repository import Gio
from gi.repository import Gtk

import pdfchain.paths as paths
import pdfchain.window as window
from pdfchain.preferences import Preferences


class Application(Gtk.Application):
    """
    """

    _preferences = None


    def __init__(self, *args, **kwargs):
        super().__init__(*args, application_id="net.sourceforge.pdfchain",
                         flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE, **kwargs)

        # Init preferences
        self._preferences = Preferences()

        # Add command line options
        self.add_main_option("test", ord("t"), GLib.OptionFlags.NONE,
                             GLib.OptionArg.NONE, "Command line test", None)


    # Overridden Class Signal Handler
    def do_startup(self):
        '''
        @brief Overwrite Virtual Class: Do Startup
        @details
            Connect the application menu actions with the application.
        @see <https://wiki.gnome.org/HowDoI/ApplicationMenu>
        '''
        Gtk.Application.do_startup(self)

        # Create and connect app menu entries
        action = Gio.SimpleAction.new("new", None)
        action.connect("activate", self.on_menu_new)
        self.add_action(action)

        action = Gio.SimpleAction.new("preferences", None)
        action.connect("activate", self.on_menu_preferences)
        self.add_action(action)

        # action = Gio.SimpleAction.new("shortcuts", None)
        # action.connect("activate", self.on_menu_shortcuts)
        # self.add_action(action)

        # action = Gio.SimpleAction.new("help", None)
        # action.connect("activate", self.on_menu_help)
        # self.add_action(action)

        action = Gio.SimpleAction.new("about", None)
        action.connect("activate", self.on_menu_about)
        self.add_action(action)

        action = Gio.SimpleAction.new("close", None)
        action.connect("activate", self.on_menu_close)
        self.add_action(action)

       # # Create app menu (deprecated)
       # builder_menu = Gtk.Builder()
       # builder_menu.add_from_file(paths.get_ui_file("menus.ui"))
       # self.set_app_menu(builder_menu.get_object("app-menu"))


    def do_activate(self):
        '''
        @brief Override Virtual Gio.Application Method.
        @details
            Appends the first application window.
        '''
        self._append_app_window()


    def do_command_line(self, command_line):
        '''
        TODO
        '''
        options = command_line.get_options_dict()
        options = options.end().unpack()

        if "test" in options:
            print("Test argument received: '%s'" % options["test"])

        self.activate()
        return 0


    # Menu Signal Handler
    def on_menu_new(self, action, param):
        '''
        @brief Application Menu Signal Handler: New Window
        @details
            Creates a new application window instance.
        @see window.AppWindowHandler.on_menu_app_new_activate()
        '''
        self._append_app_window()


    def on_menu_preferences(self, action, param):
        '''
        @brief Application Menu Signal Handler: Preferences
        @details
            Shows the preferences dialog.
        @see window.AppWindowHandler.on_menu_app_pref_activate()
        '''
        # self._preferences.set_transient_for(self.get_active_window())
        # self._preferences.run_dialog()
        self._preferences.run_dialog(self.get_active_window())


    def on_menu_shortcuts(self, action, param):
        '''
        @brief Application Menu Signal Handler: Shortcuts
        @details
            Shows the shortcuts information table dialog.
        @see window.AppWindowHandler.on_menu_app_scuts_activate()
        '''
        print("on_menu_shortcuts()") #TEST


    def on_menu_help(self, action, param):
        '''
        @brief Application Menu Signal Handler: Help
        @details
            Shows the helping manual.
        @see window.AppWindowHandler.on_menu_app_help_activate()
        '''
        print("on_menu_help()") #TEST


    def on_menu_about(self, action, param):
        '''
        @brief Application Menu Signal Handler: About PDF Chain NEO
        @details
            Shows the about dialog.
        @see window.AppWindowHandler.on_menu_app_about_activate()
        '''
        builder = Gtk.Builder()
        builder.add_from_file(paths.get_ui_file("dialogs.ui"))
        about_dialog = builder.get_object("dialog_about")
        about_dialog.set_transient_for(self.get_active_window())
        about_dialog.run()
        about_dialog.destroy()


    def on_menu_close(self, action, param):
        '''
        @brief Application Menu Signal Handler: Close Application
        @details
            Quits the application with all windows.
        @see window.AppWindowHandler.on_menu_app_close_activate()
        '''
        self.quit()


    # Methods
    def _append_app_window(self):
        '''
        Create a new ApplicaitonWindowHandler instance,
        which crates a new ApplicationWindow using the Builder from the UI file.
        It is not possible to create a second ApplicationWindow instance
        from the same Builder instance.
        A `copy.deepcopy` of the Builder object does not work.
        ("TypeError: GObject descendants' instances are non-copyable").
        If no independent builder object is used,
        all ApplicationWindow objects would point to the same Gtk widget
        in the background and no new instance will be shown.
        '''
        window.AppWindowHandler(self)



# Copyright (C) Martin Singer <martin.singer@web.de>
#
# This file is part of PDF Chain.
#
# PDF Chain is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PDF Chain is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PDF Chain.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Martin Singer

"""
PDF Chain is a graphical user interface for the PDF Toolkit (PDFtk).
The GUI supports all common features of the command line tool in a comfortable way.

PDF Chain generates a command for the PDF Toolkit from the GUI settings
and executes it on the system.
Therefore the PDF Toolkit must be already installed on the system.
"""

import datetime


__version__ = "0.5.0.0"
__author__ = "Martin Singer"
__author_email__ = "martin.singer@web.de"
__website_url__ = "pdfchain.sourceforge.net"
__license__ = "GPL v3+"
__copyright__ = u"Copyright \xa9 2009 - %d %s" % (datetime.date.today().year, __author__)


import gettext
import locale

import pdfchain.paths as paths


# def __init_translations():
#     """
#     @brief TODO
#     @details
#         TODO
#     """
#     print(__copyright__)
#
#     _GETTEXT_DOMAIN = "pdfchain"
#     _LOCALE_DIR = paths.get_locale_dir()
#
#     locale.setlocale(locale.LC_ALL, '')
#
#     for module in locale, gettext:
#         module.bindtextdomain(_GETTEXT_DOMAIN, _LOCALE_DIR)
#         module.textdomain(_GETTEXT_DOMAIN)
#
#
# __init_translations()



#!/usr/bin/env python3

# Copyright (C) Martin Singer <martin.singer@web.de>
#
# This file is part of PDF Chain.
#
# PDF Chain is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PDF Chain is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PDF Chain.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Martin Singer

'''
@brief The Application Window
@details
    This module contains classes handling the application window.

@see ./__init__.py
@see ./window.ui
'''

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk
from gi.repository import GLib

#gi.require_version('Handy', '1')
#from gi.repository import Handy

# Local Libs
from pdfchain import paths as paths
from pdfchain import cli as cli
from pdfchain import filechooser as filechooser
from pdfchain import strings as strings
from pdfchain.cmd_pdftk import PDFtk as pdftk


class AppWindowHandler:
    '''
    @brief The Application Window Handler.
    @details
        This class contains signal handlers and dependent helper methods.
        The window GUI and its signals, is defined in an external XML file.
        This file is loaded and interpreted by a builder,
        and the signals defined in it are linked to the signal handlers
        of this class.
    '''

    # Constant column and row numbers
    _COLUMN_ID          = 0  # ComboBox ListStore column
    _COLUMN_TEXT        = 1
    _ROW_FILTER_ALL     = 0  # ComboBox ListStore entry
    _ROW_ROTATION_NORTH = 0

    _COLUMN_ACTIVE        = 0  # TreeView ListStore column (cat documents)
    _COLUMN_PATH          = 1
    #_COLUMN_filename      = 2  # not required
    #_COLUMN_tooltip       = 3  # not required
    _COLUMN_SELECTION     = 4
    _COLUMN_FILTER_TEXT   = 5
    _COLUMN_ROTATION_TEXT = 6
    _COLUMN_PASSWORD      = 7
    _COLUMN_PAGES         = 8
    _COLUMN_INDEX         = 9
    _COLUMN_FILTER_ID     = 10
    _COLUMN_ROTATION_ID   = 11

    _COLUMN_ADD = 0           # TreeView ListStore column (attach files)

    # Constant strings
    _STR_PATH = "Path"  # Used for the tool tip strings

    # Variables
    _source_document_pages = 0  # FIXME: pages are not counted, when document was selected!
    #_layer_document_pages = 0  # TODO: implement this variable!
                                #       Only if document has more than one page,
                                #       multiple background / stamp is able

    # Declaration of the title and sub title strings
    # (the strings will be copied from the UI file)
    _str_title_main   = ""
    _str_title_cat    = ""
    _str_title_burst  = ""
    _str_title_layer  = ""
    _str_title_attach = ""
    _str_title_tools  = ""

    _str_subtitle_main   = ""
    _str_subtitle_cat    = ""
    _str_subtitle_burst  = ""
    _str_subtitle_layer  = ""
    _str_subtitle_attach = ""
    _str_subtitle_tools  = ""


    def __init__(self, app):
        '''
        @brief The Initialization Method of the Application Window Class.
        @details
            This method creates a local builder object, used to load the UI file
            with the application window and connects the defined signals
            with the signal handlers defined in this class.
            It also connects the application window to the application object,
            which manages application window instances.
            Using the builder object also required widgets and list store objects
            are imported from the UI file.

        @param[in] app Reference to the application object
        '''
        # Create the Builder object, load the UI file and connect the SignalHandler
        builder = Gtk.Builder()
        builder.add_from_file(paths.get_ui_file("window.ui"))
        builder.connect_signals(self)

        # Add the ApplicationWindow to the Application
        app_window = builder.get_object("applicationwindow")
        app.add_window(app_window)
        #cli.debug_message("TEST: type: {} ".format(type(app_window))) #TEST: show the object type
        #cli.debug_message("TEST: app_window: {}".formtat(app_window)) #TEST: show the memory address of the object instances

        # Import required ListStores
        self._lstore_cat_document_id = builder.get_object("lstore_cat_document_id")
        self._lstore_cat_documents = builder.get_object("lstore_cat_documents")
        self._lstore_cat_filter = builder.get_object("lstore_cat_filter")
        self._lstore_cat_rotation = builder.get_object("lstore_cat_rotation")
        #self._lstore_burst_counter_base = builder.get_object("lstore_burst_counter_base") # not required
        self._lstore_attach_files = builder.get_object("lstore_attach_files")

        # Import required widgets of the Header Bar
        self._headerbar            = builder.get_object("headerbar")
        self._revealer_button_menu = builder.get_object("revealer_button_menu")
        self._revealer_button_help = builder.get_object("revealer_button_help")
        self._stack_hbar_right     = builder.get_object("stack_hbar_right")
        self._page_appmenu         = builder.get_object("page_appmenu")
        self._page_exec            = builder.get_object("page_exec")
        self._mbutton_rest         = builder.get_object("mbutton_rest")

        # Import required widgets of the InfoBar
        self._infobar = builder.get_object("infobar")
        self._label_infobar = builder.get_object("label_infobar")

        # Import required widgets of page Main
        self._stack_main       = builder.get_object("stack_main")
        self._page_main_menu   = builder.get_object("page_main_menu")
        self._page_main_cat    = builder.get_object("page_main_cat")
        self._page_main_srcdoc = builder.get_object("page_main_srcdoc")

        self._label_main_title_cat    = builder.get_object("label_main_title_cat")
        self._label_main_title_burst  = builder.get_object("label_main_title_burst")
        self._label_main_title_layer  = builder.get_object("label_main_title_layer")
        self._label_main_title_attach = builder.get_object("label_main_title_attach")
        self._label_main_title_tools  = builder.get_object("label_main_title_tools")
        self._label_main_subtitle_cat    = builder.get_object("label_main_subtitle_cat")
        self._label_main_subtitle_burst  = builder.get_object("label_main_subtitle_burst")
        self._label_main_subtitle_layer  = builder.get_object("label_main_subtitle_layer")
        self._label_main_subtitle_attach = builder.get_object("label_main_subtitle_attach")
        self._label_main_subtitle_tools  = builder.get_object("label_main_subtitle_tools")

        # Import required widgets of page Main - Menu
        #self._listbox_main_menu = builder.get_object("listbox_main_menu") # not required
        self._lbrow_main_cat    = builder.get_object("lbrow_main_cat")
        self._lbrow_main_burst  = builder.get_object("lbrow_main_burst")
        self._lbrow_main_layer  = builder.get_object("lbrow_main_layer")
        self._lbrow_main_attach = builder.get_object("lbrow_main_attach")
        self._lbrow_main_tools  = builder.get_object("lbrow_main_tools")

        #TEST
        #self._lbrow_main_burst.set_header(Gtk.HSeparator()) # FIXME: works only in Gtk.ListBoxUpdateHeaderFunc()

        # Import required widgets of page Main - Source File
        self._fcbutton_srcdoc       = builder.get_object("fcbutton_srcdoc")
        self._stack_srcdoc          = builder.get_object("stack_srcdoc")
        self._subpage_srcdoc_attach = builder.get_object("subpage_srcdoc_attach")
        self._subpage_srcdoc_layer  = builder.get_object("subpage_srcdoc_layer")
        self._subpage_srcdoc_burst  = builder.get_object("subpage_srcdoc_burst")
        self._subpage_srcdoc_tools  = builder.get_object("subpage_srcdoc_tools")

        # Import required widgets of page Concatenate
        self._ttbutton_cat_shuffle  = builder.get_object("ttbutton_cat_shuffle")
        self._cbox_cat_document_id  = builder.get_object("cbox_cat_document_id")
        self._entry_cat_document_id = builder.get_object("entry_cat_document_id")
        self._tview_cat_documents   = builder.get_object("tview_cat_documents")

        # Import required widgets of sub page Attachments
        #self._rtbutton_attach_to_doc = builder.get_object("rtbutton_attach_to_document") # not required
        self._rtbutton_attach_page   = builder.get_object("rtbutton_attach_to_page")
        self._toolitem_attach_page   = builder.get_object("toolitem_attach_page")
        self._sbutton_attach_page    = builder.get_object("sbutton_attach_page")
        self._tview_attach_files     = builder.get_object("tview_attach_files")

        # Import required widgets of sub page Layer (Watermark / Stamp)
        self._fcbutton_layer_document = builder.get_object("fcbutton_layer_document")
        self._rbutton_layer_watermark = builder.get_object("rbutton_layer_watermark")
        #self._rbutton_layer_stamp     = builder.get_object("rbutton_layer_stamp") # not required
        self._switch_layer_multiple   = builder.get_object("switch_layer_multiple")

        # Import required widgets of sub page Burst
        self._entry_burst_prefix   = builder.get_object("entry_burst_prefix")
        self._cbox_burst_base      = builder.get_object("cbox_burst_base")
        #self._entry_burst_base     = builder.get_object("entry_burst_base")
        self._rbutton_burst_auto   = builder.get_object("rbutton_burst_auto")
        self._rbutton_burst_manual = builder.get_object("rbutton_burst_manual")
        self._sbutton_burst_manual = builder.get_object("sbutton_burst_manual")
        self._entry_burst_suffix   = builder.get_object("entry_burst_suffix")
        self._switch_burst_ext     = builder.get_object("switch_burst_extension")
        self._label_burst_template = builder.get_object("label_burst_template_preview")

        # Import required widgets of sub page Tools
        self._rbutton_tools_unpack            = builder.get_object("rbutton_tools_unpack")

        self._rbutton_tools_dump_annots       = builder.get_object("rbutton_tools_dump_annots")
        self._rbutton_tools_ddata             = builder.get_object("rbutton_tools_ddata")
        self._cbutton_tools_ddata_utf8        = builder.get_object("cbutton_tools_ddata_utf8")
        self._rbutton_tools_ddfields          = builder.get_object("rbutton_tools_ddfields")
        self._cbutton_tools_ddfields_utf8     = builder.get_object("cbutton_tools_ddfields_utf8")
        self._box_tools_update_info           = builder.get_object("box_tools_update_info")
        self._rbutton_tools_update_info       = builder.get_object("rbutton_tools_update_info")
        self._cbutton_tools_update_info_utf8  = builder.get_object("cbutton_tools_update_info_utf8")
        self._fcbutton_tools_dump_file        = builder.get_object("fcbutton_tools_dump_file")

        self._rbutton_tools_generate_fdf      = builder.get_object("rbutton_tools_generate_fdf")
        self._box_tools_fform_file            = builder.get_object("box_tools_fform_file")
        self._rbutton_tools_fform             = builder.get_object("rbutton_tools_fform")
        self._box_tools_fform_options         = builder.get_object("box_tools_fform_options")
        self._cbutton_tools_fform_flatten     = builder.get_object("cbutton_tools_fform_flatten")
        self._cbutton_tools_fform_appearances = builder.get_object("cbutton_tools_fform_need_appearances")
        self._cbutton_tools_fform_drop_xfa    = builder.get_object("cbutton_tools_fform_drop_xfa")
        self._fcbutton_tools_fdf_file         = builder.get_object("fcbutton_tools_fdf_file")
        self._rbutton_tools_flatten           = builder.get_object("rbutton_tools_flatten")
        self._rbutton_tools_drop_xfa          = builder.get_object("rbutton_tools_drop_xfa")

        self._rbutton_tools_uncompress        = builder.get_object("rbutton_tools_uncompress")
        self._rbutton_tools_compress          = builder.get_object("rbutton_tools_compress")

        self._rbutton_tools_repair            = builder.get_object("rbutton_tools_repair")

        # Import required widgets of Restrictions popover menu
        self._entry_rest_passwd_user               = builder.get_object("entry_rest_passwd_user")
        self._entry_rest_passwd_owner              = builder.get_object("entry_rest_passwd_owner")
        self._rbutton_rest_enc_none                = builder.get_object("rbutton_rest_enc_none")
        self._rbutton_rest_enc_rc4_40bit           = builder.get_object("rbutton_rest_enc_rc4_40bit")
        self._rbutton_rest_enc_rc4_128bit          = builder.get_object("rbutton_rest_enc_rc4_128bit")
        self._frame_rest_perm                      = builder.get_object("frame_rest_perm")
        self._cbutton_rest_perm_printing           = builder.get_object("cbutton_rest_perm_printing")
        self._cbutton_rest_perm_degraded_printing  = builder.get_object("cbutton_rest_perm_degraded_printing")
        self._cbutton_rest_perm_copy_contents      = builder.get_object("cbutton_rest_perm_copy_contents")
        self._cbutton_rest_perm_screen_readers     = builder.get_object("cbutton_rest_perm_screenreaders")
        self._cbutton_rest_perm_modify_contents    = builder.get_object("cbutton_rest_perm_modify_contents")
        self._cbutton_rest_perm_assembly_contents  = builder.get_object("cbutton_rest_perm_assembly_contents")
        self._cbutton_rest_perm_modify_annotations = builder.get_object("cbutton_rest_perm_modify_annotations")
        self._cbutton_rest_perm_fill_in_annots     = builder.get_object("cbutton_rest_perm_fill_annotations")

        # Add File Filter to the File Chooser Dialogs
        # FIXME: set them transient/modal (but I don't know how)
        # Maybe I should create them here, using `Gtk.FileChooserButton.new_with_dialog()`
        self._fcbutton_srcdoc.add_filter(filechooser.FileFilter_PDF())
        self._fcbutton_srcdoc.add_filter(filechooser.FileFilter_All())

        self._fcbutton_layer_document.add_filter(filechooser.FileFilter_PDF())
        self._fcbutton_layer_document.add_filter(filechooser.FileFilter_All())

        self._fcbutton_tools_dump_file.add_filter(filechooser.FileFilter_Dump())
        self._fcbutton_tools_dump_file.add_filter(filechooser.FileFilter_Text())
        self._fcbutton_tools_dump_file.add_filter(filechooser.FileFilter_All())

        self._fcbutton_tools_fdf_file.add_filter(filechooser.FileFilter_FDF())
        self._fcbutton_tools_fdf_file.add_filter(filechooser.FileFilter_Text())
        self._fcbutton_tools_fdf_file.add_filter(filechooser.FileFilter_All())

        # Instance file chooser dialogs
        self._fcdialog_open_doc_multi = filechooser.FCDialog_OpenDocMulti(app_window)
        self._fcdialog_open_all_multi = filechooser.FCDialog_OpenAllMulti(app_window)
        self._fcdialog_select_folder  = filechooser.FCDialog_SelectFolder(app_window)
        self._fcdialog_save_doc       = filechooser.FCDialog_SaveDoc(app_window)
        self._fcdialog_save_dump_file = filechooser.FCDialog_SaveDump(app_window)
        self._fcdialog_save_fdf_file  = filechooser.FCDialog_SaveFDF(app_window)

        # Read the title and subtitle strings from the widgets
        self._str_title_main   = self._headerbar.get_title()
        self._str_title_cat    = self._label_main_title_cat.get_text()
        self._str_title_burst  = self._label_main_title_burst.get_text()
        self._str_title_layer  = self._label_main_title_layer.get_text()
        self._str_title_attach = self._label_main_title_attach.get_text()
        self._str_title_tools  = self._label_main_title_tools.get_text()

        self._str_subtitle_main   = self._headerbar.get_subtitle()
        self._str_subtitle_cat    = self._label_main_subtitle_cat.get_text()
        self._str_subtitle_burst  = self._label_main_subtitle_burst.get_text()
        self._str_subtitle_layer  = self._label_main_subtitle_layer.get_text()
        self._str_subtitle_attach = self._label_main_subtitle_attach.get_text()
        self._str_subtitle_tools  = self._label_main_subtitle_tools.get_text()

        # Create the Application Menu
        mbutton_appmenu = builder.get_object("mbutton_appmenu")

        builder_appmenu = Gtk.Builder()
        builder_appmenu.add_from_file(paths.get_ui_file("menus.ui"))
        appmenu = builder_appmenu.get_object("app-menu")
        popover_appmenu = Gtk.Popover.new_from_model(mbutton_appmenu, appmenu)
        mbutton_appmenu.set_popover(popover_appmenu)

        # Show the application window and all widgets
        app_window.present()


    # Handler for Window Signals
    def on_app_window_delete_event(self, *args):
        '''Signal handler to quit the application window.'''
        Gtk.main_quit(*args)


    # Handler for the InfoBar action widgets
    def on_infobar_response(self, infobar, response_id):
        '''
        @brief Handler for the InfoBar response signal
        @details
            Hides the InfoBar when the response ID is Gtk.ResponseType.CLOSE

        @param[in] infobar     The object which received the signal
        @param[in] response_id The response ID
        '''
        if response_id == Gtk.ResponseType.CLOSE:
            self._infobar.set_revealed(False)


    def _infobar_message(self, message_type, str_message):
        '''
        @brief Method to show user messages in the infobar.
        @details
            First, this method calls a function to print the message
            on the command line.  Then it sets the message type
            and the message text of the InfoBar and then displays it.

        @see cli.message()

        @param[in] message_type The `Gtk.MessageType`
        @param[in] str_message  The message string
        '''
        cli.message(message_type, str_message)
        self._infobar.set_message_type(message_type)
        self._label_infobar.set_text(str_message)
        self._infobar.set_revealed(True)


    # Handler for the HeaderBar action widgets
    def on_button_help_clicked(self, *args):
        '''
        @brief Handler for the Help button clicked signal
        @details
            Shows the help system for the shown section.

        @TODO: not implemented yet
        '''
        self._infobar_message(Gtk.MessageType.INFO, "The help system is not implemented yet!")


    def on_button_menu_clicked(self, *args):
        '''
        @brief Signal Handler for Button Menu clicked (back button)
        @details
            Sets the window widgets for the main menu.
        @see on_listbox_main_menu_row_activated()
        '''
        self._infobar.set_revealed(False)
        self._headerbar.set_title(self._str_title_main)
        self._headerbar.set_subtitle(self._str_subtitle_main)
        self._stack_main.set_visible_child(self._page_main_menu)
        self._revealer_button_menu.set_reveal_child(False)
        self._revealer_button_help.set_reveal_child(False)
        self._stack_hbar_right.set_visible_child(self._page_appmenu)


    # Handler for all pages (execute)
    def on_button_execute_clicked(self, button):
        """
        @brief Signal Handler for the execute button ("Save Document As").
        @details
            Starts the creation and execution of the command.

        @param[in] button The object which received the signal _(not used)_
        """
        str_section_command = self._create_pdftk_cmd()
        if str_section_command:
            str_command = " ".join([pdftk.get_command(), str_section_command])
            exit_code = cli.execute(str_command)
            self._evaluate_pdftk_exit_codes(exit_code)


    def _evaluate_pdftk_exit_codes(self, exit_code):
        '''
        @brief Evaluation of the PDFtk exit codes.
        @details
            If the exit code is not 0, a message with the PDFtk exit code appears.
        @par
            The PDFtk exit codes are not unique,
            rather they should be understood in the context of the selected options
            and also the error messages output.
            This makes it very difficult and difficult to deduce error causes
            from them correctly without parsing the error messages
            - which would also be very difficult.
            For this reason, the exit codes are not evaluated.
        @todo
            A terminal display of the output messages is planned for a later version.

        @see <https://www.pdflabs.com/tools/pdftk-the-pdf-toolkit/pdftk-2.02-src.zip>

        @param[in] exit_code The PDFtk exit code
        '''
        if exit_code == 0:
            #self._infobar.set_revealed(False)
            pass
        elif exit_code == 32512:
            self._infobar_message(Gtk.MessageType.ERROR,
                    "PDF Toolkit command '{}' not found!".format(pdftk.get_command()))
        else:
            self._infobar_message(Gtk.MessageType.ERROR,
                    "PDFtk exited with error code '{}'!".format(str(exit_code)))


    def _verify_rest_passwd_enc_is_correct(self):
        '''
        @brief Verify that both restriction passwords are correctly encoded.
        @details
            Only Characters of the character set Latin-1 are allowed.
            If a illegal character was detected, a warning message is displayed.
        @see self.on_button_execute_clicked()
        @see self._create_pdftk_cmd_cat_paths()

        @retval False One of the two passwords contains illegal characters
        @retval True  Both passwords contain no illegal characters
        '''
        str_passwd = self._entry_rest_passwd_user.get_text()
        #self._entry_rest_passwd_user.get_style_context().add_class("error") #TEST
        if self._str_is_latin1_encoded(str_passwd) == False:
            self._infobar_message(Gtk.MessageType.WARNING,
                    "User password contains illegal characters!"
                    " Only Latin-1 encoding is allowed.")
            return False
        str_passwd = self._entry_rest_passwd_owner.get_text()
        if self._str_is_latin1_encoded(str_passwd) == False:
            self._infobar_message(Gtk.MessageType.WARNING,
                    "Owner password contains illegal characters!"
                    " Only Latin-1 encoding is allowed.")
            return False
        return True


    def _str_is_latin1_encoded(self, string):
        '''
        @brief Check if string is Latin-1 encoded.
        @see self._verify_rest_passwd_enc_is_correct()

        @param[in] string The string to check
        @retval False String contains illegal characters
        @retval True  String contains no illegal characters
        '''
        try:
            string.encode(encoding="latin-1", errors="strict")
        except UnicodeError:
            return False
        return True


    def _create_pdftk_cmd(self):
        '''
        @brief Creation of the PDFtk command.
        @details
            This method initiates the building of the PDFtk command.
            When it was able to create a valid command,
            the method also initiates the execution of the command.
            The creation of the command depends on the shown stack children,
            which create their command part and return it back.
            Depending on the selected stack child,
            is checked that the user and owner passwords
            do not contain illegal characters before the command is created.

        @par Stack Children
            - `_page_main_cat`
            - `_page_main_srcdoc`
              - `_subpage_srcdoc_burst`
              - `_subpage_srcdoc_layer`
              - `_subpage_srcdoc_attach`
              - `_subpage_srcdoc_tools`

        @see on_button_execute_clicked()
        @see self._create_pdftk_cmd_cat()
        @see self._create_pdftk_cmd_burst()
        @see self._create_pdftk_cmd_layer()
        @see self._create_pdftk_cmd_attach()
        @see self._create_pdftk_cmd_tools()
        @see self._create_pdftk_cmd_rest()
        '''
        str_section_command = ""

        if self._stack_main.get_visible_child() is self._page_main_cat:
            if self._verify_rest_passwd_enc_is_correct() == True:
                str_section_command = self._create_pdftk_cmd_cat()
                if str_section_command:
                    str_section_command += self._create_pdftk_cmd_rest()
        elif self._stack_main.get_visible_child() is self._page_main_srcdoc:
            source_document = self._fcbutton_srcdoc.get_filename()
            if source_document:
                str_section_options = ""
                if self._stack_srcdoc.get_visible_child() is self._subpage_srcdoc_burst:
                    if self._verify_rest_passwd_enc_is_correct() == True:
                        str_section_options = self._create_pdftk_cmd_burst()
                        if str_section_options:
                            str_section_options += self._create_pdftk_cmd_rest()
                elif self._stack_srcdoc.get_visible_child() is self._subpage_srcdoc_layer:
                    if self._verify_rest_passwd_enc_is_correct() == True:
                        str_section_options = self._create_pdftk_cmd_layer()
                        if str_section_options:
                            str_section_options += self._create_pdftk_cmd_rest()
                elif self._stack_srcdoc.get_visible_child() is self._subpage_srcdoc_attach:
                    if self._verify_rest_passwd_enc_is_correct() == True:
                        str_section_options = self._create_pdftk_cmd_attach()
                        if str_section_options:
                            str_section_options += self._create_pdftk_cmd_rest()
                elif self._stack_srcdoc.get_visible_child() is self._subpage_srcdoc_tools:
                    str_section_options = self._create_pdftk_cmd_tools()
                else:
                    cli.debug_message("WARNING: Unknown sub page!")
                    #TODO: maybe rise exception

                if str_section_options:
                    str_section_command = strings.quote_and_escape_string(source_document)
                    str_section_command += ' ' + str_section_options
            else:
                self._infobar_message(Gtk.MessageType.INFO,
                        "No source document was chosen!")
        else:
            cli.debug_message("WARNING: Unknown main page!")
            #TODO: maybe rise exception

        return str_section_command


    def _create_pdftk_cmd_cat(self):
        '''
        @brief Creation of the PDFtk concatenate command.
        @details
            This is the first part of the concatenate command creation.
            It runs the output file chooser dialog
            and evaluates the static widgets of the concatenate page.
            It also calls a method to get the concatenate path command.

        @see self.on_button_execute_clicked()
        @see self._create_pdftk_cmd_cat_paths()

        @return str_command String with the concatenate part of the PDFtk command
        '''
        str_command = ""

        str_cat_paths = self._create_pdftk_cmd_cat_paths()
        if str_cat_paths: #TODO: check here, if iter_first is not None, and read paths when output doc is selected
            str_target = filechooser.fchooser_run(self._fcdialog_save_doc)
            if str_target:
                str_document_id = ""
                treeiter_cbox_doc_id = self._cbox_cat_document_id.get_active_iter()
                if treeiter_cbox_doc_id is None:
                    str_document_id = pdftk.ID_KEEP
                else:
                    str_active_id = self._lstore_cat_document_id.get_value(
                            treeiter_cbox_doc_id, self._COLUMN_ID)
                    if str_active_id == "new":
                        str_document_id = pdftk.ID_NEW
                    elif str_active_id == "first":
                        str_document_id = pdftk.ID_FIRST
                    elif str_active_id == "final":
                        str_document_id = pdftk.ID_FINAL
                    else:
                        cli.debug_message("WARNING: Unknown document ID")

                str_command += str_cat_paths
                str_command += ' ' + pdftk.OUTPUT
                str_command += ' ' + strings.quote_and_escape_string(str_target)
                if str_document_id:
                    str_command += ' ' + str_document_id
#        else:
#            self._infobar_message(Gtk.MessageType.INFO, "No active document for concatenation was found in the list!")

        return str_command


    def _create_pdftk_cmd_cat_paths(self):
        '''
        @brief Creation of the PDFtk concatenate path command.
        @details
            This is the second part of the creation of the concatenate command.
            It evaluates the concatenate list store.

        @see self._create_pdftk_cmd_cat()

        @return str_command String with the second part of the concatenate command
        '''
        str_command = ""
        self._index_cat_docs() #TODO: Check later, if this call is required here anymore

        # Append handles for every individual document stored in an active entry
        num_index = 0
        treeiter_entry = self._lstore_cat_documents.get_iter_first()
        while treeiter_entry is not None:
            if self._lstore_cat_documents.get_value(treeiter_entry, self._COLUMN_INDEX) == num_index:
                str_path = self._lstore_cat_documents.get_value(treeiter_entry, self._COLUMN_PATH)
                str_command += ' ' + strings.create_handle_from_number(num_index)
                str_command += '=' + strings.quote_and_escape_string(str_path)
                num_index += 1
            treeiter_entry = self._lstore_cat_documents.iter_next(treeiter_entry)

        if not str_command:
            self._infobar_message(Gtk.MessageType.INFO,
                    "No active document for concatenation was found in the list!")
            return "" # return empty command string

        # Append input password command, if at least one entry with a password was found
        treeiter_entry = self._lstore_cat_documents.get_iter_first()
        while treeiter_entry is not None:
            if self._lstore_cat_documents.get_value(treeiter_entry, self._COLUMN_PASSWORD):
                str_command += ' ' + pdftk.PASSWORD_INPUT
                break
            treeiter_entry = self._lstore_cat_documents.iter_next(treeiter_entry)

        # Assign input passwords to their corresponding handle
        num_index = 0
        treeiter_entry = self._lstore_cat_documents.get_iter_first()
        while treeiter_entry is not None:
            if self._lstore_cat_documents.get_value(treeiter_entry, self._COLUMN_INDEX) == num_index:
                str_password = self._lstore_cat_documents.get_value(treeiter_entry, self._COLUMN_PASSWORD)
                if str_password:
                    if self._str_is_latin1_encoded(str_password) == False:
                        self._infobar_message(Gtk.MessageType.INFO,
                                "Input password contains illegal characters!"
                                " Only Latin-1 encoding is allowed.")
                        return "" # return empty command string
                    str_command += ' ' + strings.create_handle_from_number(num_index)
                    str_command += '=' + strings.quote_and_escape_string(str_password)
            num_index += 1
            treeiter_entry = self._lstore_cat_documents.iter_next(treeiter_entry)

        # Insert the command "cat" or "shuffle"
        if self._ttbutton_cat_shuffle.get_active() == True:
            str_command += ' ' + pdftk.SHUFFLE
        else:
            str_command += ' ' + pdftk.CAT

        # Parsing the page selection string, selection filter and rotation
        # and appending the created command
        treeiter_entry = self._lstore_cat_documents.get_iter_first()
        while treeiter_entry is not None:
            num_index = self._lstore_cat_documents.get_value(treeiter_entry, self._COLUMN_INDEX)
            if num_index < GLib.MAXUINT:
                str_selection = self._lstore_cat_documents.get_value(treeiter_entry, self._COLUMN_SELECTION)
                valid, str_selection = strings.validate_page_selection_string(str_selection)
                if valid == True:

                    str_filter_id = self._lstore_cat_documents.get_value(treeiter_entry, self._COLUMN_FILTER_ID)
                    if str_filter_id == 'even':
                        str_filter = pdftk.PAGES_EVEN
                    elif str_filter_id == 'odd':
                        str_filter = pdftk.PAGES_ODD
                    else:
                        str_filter = pdftk.PAGES_ALL

                    str_rotation_id = self._lstore_cat_documents.get_value(treeiter_entry, self._COLUMN_ROTATION_ID)
                    if str_rotation_id == 'east':
                        str_rotation = pdftk.ROTATION_EAST
                    elif str_rotation_id == 'south':
                        str_rotation = pdftk.ROTATION_SOUTH
                    elif str_rotation_id == 'west':
                        str_rotation = pdftk.ROTATION_WEST
                    else:
                        str_rotation = pdftk.ROTATION_NORTH

                    arr_selection_parts = str_selection.split(None) # consecutive whitespace are regarded as a single separator
                    for str_selection_part in arr_selection_parts:
                        str_command += ' ' + strings.create_handle_from_number(num_index)
                        str_command += str_selection_part
                        str_command += str_filter
                        str_command += str_rotation
                else:
                    self._infobar_message(Gtk.MessageType.INFO, "Invalid page selection: '{}'".format(str_selection))
                    return "" # return empty command string

            treeiter_entry = self._lstore_cat_documents.iter_next(treeiter_entry)

        return str_command


    def _create_pdftk_cmd_burst(self):
        '''
        @brief Method to create the PDFtk command for sub page Burst.
        @details
            This method creates the PDFtk command
            by evaluating the widgets of the Burst sub page.

        @see self.on_button_execute_clicked()

        @return str_command String with the burst command
        '''
        str_command = ""
        str_target = filechooser.fchooser_run(self._fcdialog_select_folder)

        if str_target:
            str_template = self._label_burst_template.get_text()
            str_pattern = str_target + GLib.DIR_SEPARATOR_S + str_template

            str_command  =       pdftk.BURST
            str_command += ' ' + pdftk.OUTPUT
            str_command += ' ' + strings.quote_and_escape_string(str_pattern)

        return str_command


    def _create_pdftk_cmd_layer(self):
        '''
        @brief Method to create the PDFtk command for sub page Background / Stamp.
        @details
            This method creates the PDFtk command
            by evaluating the widgets of the Background / Stamp sub page.

        @see self.on_button_execute_clicked()

        @return str_command String with the background or stamp command
        '''
        str_command = ""
        str_layer_doc = self._fcbutton_layer_document.get_filename()
        if not str_layer_doc:
            self._infobar_message(Gtk.MessageType.INFO, "No layer document was chosen!")
        else:
            str_target = filechooser.fchooser_run(self._fcdialog_save_doc)
            if str_target:
                if self._rbutton_layer_watermark.get_active() == True:
                    if self._switch_layer_multiple.get_active() == True:
                        str_command = pdftk.BACKGROUND_MULTI
                    else:
                        str_command = pdftk.BACKGROUND
                else:
                    if self._switch_layer_multiple.get_active() == True:
                        str_command = pdftk.STAMP_MULTI
                    else:
                        str_command = pdftk.STAMP

                str_command += ' ' + strings.quote_and_escape_string(str_layer_doc)
                str_command += ' ' + pdftk.OUTPUT
                str_command += ' ' + strings.quote_and_escape_string(str_target)

        return str_command


    def _create_pdftk_cmd_attach(self):
        '''
        @brief Method to create the PDFtk command for sub page Attachments.
        @details
            This method creates the PDFtk command
            by evaluating the widgets of the Attachments sub page.
            It calls a method to get the part of the command
            with the paths from the attachment list store.

        @see self.on_button_execute_clicked()
        @see self._create_pdftk_cmd_attach_paths()

        @return str_command String with the attachment command
        '''
        str_command = ""
        str_command_options = ""
        str_attach_paths = self._create_pdftk_cmd_attach_paths()
        if str_attach_paths:
            str_target = filechooser.fchooser_run(self._fcdialog_save_doc)
            if str_target:
                str_command = pdftk.ATTACH_FILES
                str_command += str_attach_paths

                if self._rtbutton_attach_page.get_active() == True:
                    str_command_options += ' ' + pdftk.TO_PAGE
                    str_command_options += ' ' + str(int(self._sbutton_attach_page.get_value()))

                str_command += str_command_options
                str_command += ' ' + pdftk.OUTPUT
                str_command += ' ' + strings.quote_and_escape_string(str_target)
        else:
            self._infobar_message(Gtk.MessageType.INFO, "No file to attach was found in the list!")

        return str_command


    def _create_pdftk_cmd_attach_paths(self):
        '''
        @brief Method to create the PDFtk command for sub page Attachments.
        @details
            This method creates the part
            with the paths from the attachment list store
            of the Attachment sub page for the PDFtk command.
            The created command part is just a list with the files to attach.

        @see self._create_pdftk_cmd_attach()

        @return str_command String with the path part of the attachment command
        '''
        str_command = ""

        treeiter_files = self._lstore_attach_files.get_iter_first()
        while treeiter_files is not None:
            path = self._lstore_attach_files.get_value(treeiter_files, self._COLUMN_PATH)
            str_command += ' ' + strings.quote_and_escape_string(path)
            treeiter_files = self._lstore_attach_files.iter_next(treeiter_files)

        return str_command


    def _create_pdftk_cmd_tools(self):
        '''
        @brief Method to create the PDFtk command for sub page Tools.
        @details
            This method creates the PDFtk command
            by evaluating the widgets of the Tools sub page.
        @note
            If the target string is empty,
            no command string is created and also remains empty.
        @see self.on_button_execute_clicked()
        @return str_command String with the selected tool command
        '''
        # Get the target string (document, file or directory path)
        str_dump_file = self._fcbutton_tools_dump_file.get_filename()
        str_fdf_file = self._fcbutton_tools_fdf_file.get_filename()
        str_target = ""

        if self._rbutton_tools_unpack.get_active() == True:
            response = self._fcdialog_select_folder.run()
            self._fcdialog_select_folder.hide()
            if response == Gtk.ResponseType.OK:
                str_target = self._fcdialog_select_folder.get_current_folder()
        elif self._rbutton_tools_dump_annots.get_active() == True \
                or self._rbutton_tools_ddfields.get_active() == True \
                or self._rbutton_tools_ddata.get_active() == True:
            str_target = filechooser.fchooser_run(self._fcdialog_save_dump_file)
        elif self._rbutton_tools_generate_fdf.get_active() == True:
            str_target = filechooser.fchooser_run(self._fcdialog_save_fdf_file)
        else:
            if self._rbutton_tools_update_info.get_active() == True:
                if str_dump_file:
                    str_target = filechooser.fchooser_run(self._fcdialog_save_doc)
                else:
                    self._infobar_message(Gtk.MessageType.INFO,
                            "No metadata dump file was chosen! (*.dump)")
                    return ""
            elif self._rbutton_tools_fform.get_active() == True:
                if str_fdf_file:
                    str_target = filechooser.fchooser_run(self._fcdialog_save_doc)
                else:
                    self._infobar_message(Gtk.MessageType.INFO,
                            "No fill form file was chosen! (*.fdf, *.xfdf)")
                    return ""
            else:
                str_target = filechooser.fchooser_run(self._fcdialog_save_doc)
        #cli.debug_message("TEST: str_target: '{}'".format(str_target)) #TEST

        # Build the command string
        str_command = ""
        if str_target:
            str_target = strings.quote_and_escape_string(str_target)

            if self._rbutton_tools_repair.get_active() == True:
                str_command += " " + pdftk.OUTPUT
                str_command += " " + str_target
            elif self._rbutton_tools_unpack.get_active() == True:
                str_command += " " + pdftk.UNPACK_FILES
                str_command += " " + pdftk.OUTPUT
                str_command += " " + str_target
            elif self._rbutton_tools_uncompress.get_active() == True:
                str_command += " " + pdftk.OUTPUT
                str_command += " " + str_target
                str_command += " " + pdftk.UNCOMPRESS
            elif self._rbutton_tools_compress.get_active() == True:
                str_command += " " + pdftk.OUTPUT
                str_command += " " + str_target
                str_command += " " + pdftk.COMPRESS
            elif self._rbutton_tools_dump_annots.get_active() == True:
                str_command += " " + pdftk.DUMP_DATA_ANNOTS
                str_command += " " + pdftk.OUTPUT
                str_command += " " + str_target
            elif self._rbutton_tools_ddfields.get_active() == True:
                if self._cbutton_tools_ddfields_utf8.get_active() == True:
                    str_command += " " + pdftk.DUMP_DATA_FIELDS_UTF8
                else:
                    str_command += " " + pdftk.DUMP_DATA_FIELDS
                str_command += " " + pdftk.OUTPUT
                str_command += " " + str_target
            elif self._rbutton_tools_ddata.get_active() == True:
                if self._cbutton_tools_ddata_utf8.get_active() == True:
                    str_command += " " + pdftk.DUMP_DATA_UTF8
                else:
                    str_command += " " + pdftk.DUMP_DATA
                str_command += " " + pdftk.OUTPUT
                str_command += " " + str_target
            elif self._rbutton_tools_update_info.get_active() == True:
                if str_dump_file:
                    if self._cbutton_tools_update_info_utf8.get_active() == True:
                        str_command += " " + pdftk.UPDATE_INFO_UTF8
                    else:
                        str_command += " " + pdftk.UPDATE_INFO
                    str_command += " " + strings.quote_and_escape_string(str_dump_file)
                    str_command += " " + pdftk.OUTPUT
                    str_command += " " + str_target
                else:  # Has already been checked and should never be the case here
                    self._infobar_message(Gtk.MessageType.INFO,
                            "No metadata dump file was chosen! (*.dump)")
                    return ""
            elif self._rbutton_tools_generate_fdf.get_active() == True:
                str_command += " " + pdftk.GENERATE_FDF
                str_command += " " + pdftk.OUTPUT
                str_command += " " + str_target
            elif self._rbutton_tools_fform.get_active() == True:
                if str_fdf_file:
                    str_command += " " + pdftk.FILL_FORM
                    str_command += " " + strings.quote_and_escape_string(str_fdf_file)
                    str_command += " " + pdftk.OUTPUT
                    str_command += " " + str_target
                    if self._cbutton_tools_fform_drop_xfa.get_active() == True:
                        str_command += " " + pdftk.DROP_XFA
                    if self._cbutton_tools_fform_appearances.get_active() == True:
                        str_command += " " + pdftk.NEED_APPEARANCES
                    if self._cbutton_tools_fform_flatten.get_active() == True:
                        str_command += " " + pdftk.FLATTEN
                else:  # Has already been checked and should never be the case here
                    self._infobar_message(Gtk.MessageType.INFO,
                            "No fill form file was chosen! (*.fdf, *.xfdf)")
                    return ""
            elif self._rbutton_tools_flatten.get_active() == True:
                str_command += " " + pdftk.OUTPUT
                str_command += " " + str_target
                str_command += " " + pdftk.FLATTEN
            elif self._rbutton_tools_drop_xfa.get_active() == True:
                str_command += " " + pdftk.OUTPUT
                str_command += " " + str_target
                str_command += " " + pdftk.DROP_XFA

        #cli.debug_message("TEST: str_command: '{}'".format(str_command)) #TEST
        return str_command


    def _create_pdftk_cmd_rest(self):
        '''
        @brief   Method: Create the restriction options for the PDFtk command.
        @details
            This method creates a PDFtk command part
            by evaluating the widgets of the Permissions menu.

        @see self.on_button_execute_clicked()

        @return String with the restriction options for the PDFtk command

        @par When checking if one or more options are active _(OR)_
            - when 'Screen readers'      is not active, also 'Copy Contents'      is not active
            - when 'Assembly Contents'   is not active, also 'Modify Contents'    is not active
            - when 'Fill In Annotations' is not active, also 'Modify Annotations' is not active

        @par When checking if all options are active _(AND)_
            - when 'Copy Contents'      is active, also 'Screen readers'      is active
            - when 'Modify Contents'    is active, also 'Assembly Contents'   is active
            - when 'Modify Annotations' is active, also 'Fill In Annotations' is active
        '''
        str_cmd_password = ""
        str_cmd_restrictions = ""
        str_cmd_encryption = ""

        # Read Passwords
        str_passwd_user = self._entry_rest_passwd_user.get_text()
        str_passwd_owner = self._entry_rest_passwd_owner.get_text()

        if str_passwd_user or str_passwd_owner:
            if str_passwd_user == str_passwd_owner:
                # User and owner password are identical!
                # In this case the document can only be protected with the user password.
                self._infobar_message(Gtk.MessageType.WARNING,
                        "User and owner password are identical!")
                str_cmd_password += " " + pdftk.PASSWORD_USER
                str_cmd_password += " " + strings.quote_and_escape_string(str_passwd_user)
            else:
                if str_passwd_user:
                    str_cmd_password += " " + pdftk.PASSWORD_USER
                    str_cmd_password += " " + strings.quote_and_escape_string(str_passwd_user)

                if str_passwd_owner:
                    str_cmd_password += " " + pdftk.PASSWORD_OWNER
                    str_cmd_password += " " + strings.quote_and_escape_string(str_passwd_owner)

        # Read Restrictions
        if self._rbutton_rest_enc_none.get_active() == False:
            if self._cbutton_rest_perm_printing.get_active() == True \
                    or self._cbutton_rest_perm_degraded_printing.get_active() == True \
                    or self._cbutton_rest_perm_screen_readers.get_active() == True \
                    or self._cbutton_rest_perm_assembly_contents.get_active() == True \
                    or self._cbutton_rest_perm_fill_in_annots.get_active() == True:

                str_cmd_restrictions += " " + pdftk.ALLOW

                if self._cbutton_rest_perm_printing.get_active() == True \
                        and self._cbutton_rest_perm_degraded_printing.get_active() == True \
                        and self._cbutton_rest_perm_copy_contents.get_active() == True \
                        and self._cbutton_rest_perm_modify_contents.get_active() == True \
                        and self._cbutton_rest_perm_modify_annotations.get_active() == True:

                    str_cmd_restrictions += " " + pdftk.FEATURES_ALL
                else:
                    if self._cbutton_rest_perm_printing.get_active() == True:
                        str_cmd_restrictions += " " + pdftk.PRINTING

                    if self._cbutton_rest_perm_degraded_printing.get_active() == True:
                        str_cmd_restrictions += " " + pdftk.DEGRADED_PRINTING

                    if self._cbutton_rest_perm_copy_contents.get_active() == True:
                        str_cmd_restrictions += " " + pdftk.COPY_CONTENTS
                    elif self._cbutton_rest_perm_screen_readers.get_active() == True:
                        str_cmd_restrictions += " " + pdftk.SCREEN_READERS

                    if self._cbutton_rest_perm_modify_contents.get_active() == True:
                        str_cmd_restrictions += " " + pdftk.MODIFY_CONTENTS
                    elif self._cbutton_rest_perm_assembly_contents.get_active() == True:
                        str_cmd_restrictions += " " + pdftk.ASSEMBLY

                    if self._cbutton_rest_perm_modify_annotations.get_active() == True:
                        str_cmd_restrictions += " " + pdftk.MODIFY_ANNOTATIONS
                    elif self._cbutton_rest_perm_fill_in_annots.get_active() == True:
                        str_cmd_restrictions += " " + pdftk.FILL_IN

        # Read Encryption
        if self._rbutton_rest_enc_rc4_40bit.get_active() == True:
            str_cmd_encryption = " " + pdftk.ENCRYPT_40BIT
        elif self._rbutton_rest_enc_rc4_128bit.get_active() == True:
            str_cmd_encryption = " " + pdftk.ENCRYPT_128BIT
        else:
            if str_passwd_user or str_passwd_user:
                self._infobar_message(Gtk.MessageType.WARNING, "A password without an encryption has no effect!")

            if str_cmd_restrictions:
                self._infobar_message(Gtk.MessageType.WARNING, "Restrictions without an encryption have no effect!")

        # Create Permissions command
        str_command = str_cmd_password + str_cmd_restrictions + str_cmd_encryption
        return str_command


    # Handler for main page Menu
    def on_lbox_main_menu_row_activated(self, list_box, row):
        '''
        @brief Signal Handler when a ListBox row of the Main Menu was activated.
        @details
            This signal hander switches the stack pages according to the selected row.
            It also sets related widgets visible.

        @TODO Rework this method!

        @param[in] list_box The object which received the signal _(not used)_
        @param[in] row      The activated row
        '''
        if row is self._lbrow_main_cat:
            self._headerbar.set_title(self._str_title_cat)
            self._headerbar.set_subtitle(self._str_subtitle_cat)
            self._stack_main.set_visible_child(self._page_main_cat)
            self._revealer_button_menu.set_reveal_child(True)
            self._revealer_button_help.set_reveal_child(True)
            self._mbutton_rest.set_visible(True)
            self._stack_hbar_right.set_visible_child(self._page_exec)
        elif row is self._lbrow_main_burst:
            self._headerbar.set_title(self._str_title_burst)
            self._headerbar.set_subtitle(self._str_subtitle_burst)
            self._stack_srcdoc.set_visible_child(self._subpage_srcdoc_burst)
            self._stack_main.set_visible_child(self._page_main_srcdoc)
            self._revealer_button_menu.set_reveal_child(True)
            self._revealer_button_help.set_reveal_child(True)
            self._mbutton_rest.set_visible(True)
            self._stack_hbar_right.set_visible_child(self._page_exec)
            self.on_section_burst_changed()
        elif row is self._lbrow_main_layer:
            self._headerbar.set_title(self._str_title_layer)
            self._headerbar.set_subtitle(self._str_subtitle_layer)
            self._stack_srcdoc.set_visible_child(self._subpage_srcdoc_layer)
            self._stack_main.set_visible_child(self._page_main_srcdoc)
            self._revealer_button_menu.set_reveal_child(True)
            self._revealer_button_help.set_reveal_child(True)
            self._mbutton_rest.set_visible(True)
            self._stack_hbar_right.set_visible_child(self._page_exec)
        elif row is self._lbrow_main_attach:
            self._headerbar.set_title(self._str_title_attach)
            self._headerbar.set_subtitle(self._str_subtitle_attach)
            self._stack_srcdoc.set_visible_child(self._subpage_srcdoc_attach)
            self._stack_main.set_visible_child(self._page_main_srcdoc)
            self._revealer_button_menu.set_reveal_child(True)
            self._revealer_button_help.set_reveal_child(True)
            self._mbutton_rest.set_visible(True)
            self._stack_hbar_right.set_visible_child(self._page_exec)
        elif row is self._lbrow_main_tools:
            self._headerbar.set_title(self._str_title_tools)
            self._headerbar.set_subtitle(self._str_subtitle_tools)
            self._stack_srcdoc.set_visible_child(self._subpage_srcdoc_tools)
            self._stack_main.set_visible_child(self._page_main_srcdoc)
            self._revealer_button_menu.set_reveal_child(True)
            self._revealer_button_help.set_reveal_child(True)
            self._mbutton_rest.set_visible(False)
            self._stack_hbar_right.set_visible_child(self._page_exec)
        else:
            cli.debug_message("WARNING: unknown list box row: '{}'".format(row))


    # Handler for main page Concatenate
    def on_toolbutton_cat_add_doc_clicked(self, tbutton):
        '''
        @brief Signal Handler for the ToolButton Concatenate click event.
        @details
            This signal handler runs a FileChooserDialog with multiple selection
            to add new documents to the Concatenate ListStore.
            If a file name that already exists in the list is added,
            file-specific values are taken over for the new entry.
            After adding the documents,
            a method is called which re-indexes all list entries.

        @see self._get_data_from_cat_list()
        @see self._update_cat_combobox_document_id()

        @param[in] tbutton The object which received the signal _(not used)_
        '''
        response = self._fcdialog_open_doc_multi.run()
        self._fcdialog_open_doc_multi.hide()

        if response == Gtk.ResponseType.OK:
            str_filenames = self._fcdialog_open_doc_multi.get_filenames()

            treeselection = self._tview_cat_documents.get_selection()
            treemodel, treepaths = treeselection.get_selected_rows()

            if treepaths == []:
                treeiter_selected = None
            else:
                treeiter_selected = treemodel.get_iter(treepaths[0]) # get the first iterator

            for str_filename in str_filenames:
                str_doc_basename = GLib.path_get_basename(str_filename)
                str_tooltip = self._STR_PATH + ": '" + str_filename + "'"
                str_page_selection = ""
                str_page_filter = self._lstore_cat_filter[self._ROW_FILTER_ALL][self._COLUMN_TEXT]
                str_page_rotation = self._lstore_cat_rotation[self._ROW_ROTATION_NORTH][self._COLUMN_TEXT]
                str_doc_passwd = self._get_data_from_cat_list(str_filename, self._COLUMN_PASSWORD)
                str_doc_pages = self._get_data_from_cat_list(str_filename, self._COLUMN_PAGES)
                num_doc_index = GLib.MAXUINT
                str_page_filter_id = self._lstore_cat_filter[self._ROW_FILTER_ALL][self._COLUMN_ID]
                str_page_rotation_id = self._lstore_cat_rotation[self._ROW_ROTATION_NORTH][self._COLUMN_ID]

                if not str_doc_pages:
                    str_doc_pages = str(strings.count_pages_of_document(str_filename))

                if treeiter_selected == None:
                    treeiter_new = self._lstore_cat_documents.append(None)
                else:
                    treeiter_new = self._lstore_cat_documents.insert_before(
                            treeiter_selected)

                self._lstore_cat_documents[treeiter_new] = [
                        True,
                        str_filename,
                        str_doc_basename,
                        str_tooltip,
                        str_page_selection,
                        str_page_filter,
                        str_page_rotation,
                        str_doc_passwd,
                        str_doc_pages,
                        num_doc_index,
                        str_page_filter_id,
                        str_page_rotation_id]

                self._update_cat_combobox_document_id()


    def on_toolbutton_cat_remove_doc_clicked(self, tbutton):
        '''
        @brief Handler for the ToolButton Concatenate remove document clicked signal.
        @details
            This signal handler method removes selected entries
            from the Concatenate ListStore.

        @param[in] tbutton The object which received the signal _(not used)_
        '''
        treeselection = self._tview_cat_documents.get_selection()
        tree_model_selected, tree_paths_selected = treeselection.get_selected_rows()
        treeiters_selected = []
        for tree_path_selected in tree_paths_selected:
            treeiters_selected.append(tree_model_selected.get_iter(tree_path_selected))
        for treeeiter_entries_selected in treeiters_selected:
            self._lstore_cat_documents.remove(treeeiter_entries_selected)
        self._update_cat_combobox_document_id()


    def on_toolbutton_cat_duplicate_doc_clicked(self, tbutton):
        '''
        @brief Handler for the ToolButton Concatenate duplicate document clicked signal.
        @details
            This signal handler method duplicates selected entries
            in the Concatenate ListStore.

        @param[in] tbutton The object which received the signal _(not used)_
        '''
        treeselection = self._tview_cat_documents.get_selection()
        tree_model_selected, tree_paths_selected = treeselection.get_selected_rows()

        tree_iters_selected = []
        for tree_path_selected in tree_paths_selected:
            tree_iters_selected.append(tree_model_selected.get_iter(tree_path_selected))
        for treeiter_selected in tree_iters_selected:
            treeiter_new = self._lstore_cat_documents.append(None)
            self._lstore_cat_documents[treeiter_new][:] = \
                    self._lstore_cat_documents[treeiter_selected][:]


    def on_crenderer_cat_active_toggled(self, crtoggle, str_tview_path):
        '''
        @brief Signal Handler for a CellRenderer Toggle in section Cat: The Active button was toggled.
        @details
            Toggles the boolean value in the concatenate list store.

        @param[in] crtoggle       The object which received the signal _(not used)_
        @param[in] str_tview_path String representation of Gtk.TreePath describing the event location
        '''
        treeiter = self._lstore_cat_documents.get_iter(str_tview_path)
        value    = self._lstore_cat_documents.get_value(treeiter, self._COLUMN_ACTIVE)
        self._lstore_cat_documents.set_value(treeiter, self._COLUMN_ACTIVE, not value)
        self._update_cat_combobox_document_id()


    def on_crenderer_cat_selection_edited(
            self, crtext, str_tview_path, str_new_text):
        '''
        @brief Signal Handler for a CellRenderer Text in section Cat: The Page Selection was changed.
        @details
            Copy the new selection string for this entry
            into the selection string cell of the concatenate list store.

        @param[in] crtext         The object which received the signal _(not used)_
        @param[in] str_tview_path String representation of Gtk.TreePath describing the event location
        @param[in] str_new_text   The new text
        '''
        treeiter_entry = self._lstore_cat_documents.get_iter(str_tview_path)
        self._lstore_cat_documents.set_value(treeiter_entry, self._COLUMN_SELECTION, str_new_text)


    def on_crenderer_cat_filter_changed(
            self, crcombo, str_tview_path, treeiter_combo):
        '''
        @brief Signal Handler for a CellRenderer Combo in section Cat: The Page Filter was changed.
        @details
            Copies the selected text from the list store of the cell renderer combo
            into the corresponding cell of the concatenate list store.

        @param[in] crcombo        The object which received the signal _(not used)_
        @param[in] str_tview_path String representation of Gtk.TreePath describing the event location
        @param[in] treeiter_combo The new iter selected in the combo box (combo box model)
        '''
        str_text = self._lstore_cat_filter.get_value(treeiter_combo, self._COLUMN_TEXT)
        str_id = self._lstore_cat_filter.get_value(treeiter_combo, self._COLUMN_ID)
        treeiter_entry = self._lstore_cat_documents.get_iter(str_tview_path)
        self._lstore_cat_documents.set_value(treeiter_entry, self._COLUMN_FILTER_TEXT, str_text)
        self._lstore_cat_documents.set_value(treeiter_entry, self._COLUMN_FILTER_ID, str_id)


    def on_crenderer_cat_rotation_changed(
            self, crcombo, str_tview_path, treeiter_combo):
        '''
        @brief Signal Handler for a CellRenderer Combo in section Cat: The Rotation was changed.
        @details
            Copies the selected text from the list store of the cell renderer combo
            into the corresponding cell of the concatenate list store.

        @param[in] crcombo        The object which received the signal _(not used)_
        @param[in] str_tview_path String representation of Gtk.TreePath describing the event location
        @param[in] treeiter_combo The new iter selected in the combo box (combo box model)
        '''
        str_text = self._lstore_cat_rotation.get_value(treeiter_combo, self._COLUMN_TEXT)
        str_id = self._lstore_cat_rotation.get_value(treeiter_combo, self._COLUMN_ID)
        treeiter_entry = self._lstore_cat_documents.get_iter(str_tview_path)
        self._lstore_cat_documents.set_value(treeiter_entry, self._COLUMN_ROTATION_TEXT, str_text)
        self._lstore_cat_documents.set_value(treeiter_entry, self._COLUMN_ROTATION_ID, str_id)


    def on_crenderer_cat_password_edited(
            self, cellrenderer_text, str_tview_path, str_new_text):
        '''
        @brief Signal Handler for a CellRenderer Text in section Cat: The Password was edited.
        @details
            Copy the new password string for this entry
            into the password string cell of the concatenate list store.
            Then call a method to clone the password string of this entry
            into the password string cell of every entry with the same filename.

        @param[in] cellrenderer_text The object which received the signal _(not used)_
        @param[in] str_tview_path    String representation of `Gtk.TreePath describing the event location
        @param[in] str_new_text      The new text
        '''
        treeiter_entry = self._lstore_cat_documents.get_iter(str_tview_path)
        self._lstore_cat_documents.set_value(treeiter_entry,
                self._COLUMN_PASSWORD, str_new_text)
        self._sync_cat_passwords(treeiter_entry)


    def _index_cat_docs(self):
        '''
        @brief Index concatenate documents.
        @details
            This method indexes all active entries of the Concatenate list store
            with a unique ID number.
            Individual documents are identified by their path.
            Entries with the same path are assigned the same ID number.
            A index number with Glib.MAXUINT marks an inactive entry.
        '''
        # Set the index values of all entries to an "invalid" number (GLib.MAXUINT),
        # which indicates later, that this entry has no "valid" index yet.
        treeiter_A = self._lstore_cat_documents.get_iter_first()
        while treeiter_A is not None:
            self._lstore_cat_documents.set_value(treeiter_A, self._COLUMN_INDEX, GLib.MAXUINT)
            treeiter_A = self._lstore_cat_documents.iter_next(treeiter_A)

        # Search for the first unindexed entry and assign a new ID number.
        # Then continue searching for unindexed entries with the same path
        # and assign them the same ID number.
        #  - treeiter_A: compares index entries
        #  - treeiter_B: compares path  entries
        index_number = 0
        treeiter_A = treeiter_B = self._lstore_cat_documents.get_iter_first()
        while treeiter_A is not None:
            if self._lstore_cat_documents.get_value(treeiter_A, self._COLUMN_ACTIVE) == True:
                if self._lstore_cat_documents.get_value(treeiter_A, self._COLUMN_INDEX) == GLib.MAXUINT:
                    path_A = self._lstore_cat_documents.get_value(treeiter_A, self._COLUMN_PATH)
                    treeiter_B = treeiter_A
                    while treeiter_B is not None:
                        if self._lstore_cat_documents.get_value(treeiter_B, self._COLUMN_ACTIVE) == True:
                            path_B = self._lstore_cat_documents.get_value(treeiter_B, self._COLUMN_PATH)
                            if path_A == path_B:
                                self._lstore_cat_documents.set_value(treeiter_B, self._COLUMN_INDEX, index_number)
                        treeiter_B = self._lstore_cat_documents.iter_next(treeiter_B)
                    index_number += 1
            treeiter_A = self._lstore_cat_documents.iter_next(treeiter_A)

        # TEST
        #treeiter = self._lstore_cat_documents.get_iter_first()
        #while treeiter is not None:
        #    cli.debug_message("TEST: index: {}, path: '{}'".format(
        #        self._lstore_cat_documents.get_value(treeiter, self._COLUMN_INDEX),
        #        self._lstore_cat_documents.get_value(treeiter, self._COLUMN_PATH)))
        #    treeiter = self._lstore_cat_documents.iter_next(treeiter)


    def _sync_cat_passwords(self, treeiter_source):
        '''
        @brief Method: Synchronize Concatenate Passwords.
        @details
            Find row entries with the same filename like the source row (iter)
            in the concatenate list store
            and copy the password from the source row to the found rows.

        @TODO maybe using _index_cat_docs()

        param[in] treeiter_source Tree iterator pointing to the source entry
        '''
        path     = self._lstore_cat_documents.get_value(treeiter_source, self._COLUMN_PATH)
        password = self._lstore_cat_documents.get_value(treeiter_source, self._COLUMN_PASSWORD)

        treeiter_target = self._lstore_cat_documents.get_iter_first()
        while treeiter_target is not None:
            if self._lstore_cat_documents.get_value(treeiter_target, 1) == path:
                 self._lstore_cat_documents.set_value(
                         treeiter_target, self._COLUMN_PASSWORD, password)
            treeiter_target = self._lstore_cat_documents.iter_next(treeiter_target)


    def _get_data_from_cat_list(self, str_path, num_column):
        '''
        @brief   Get stored data of a document from concatenate list store.
        @details
            This method finds the first entry with the same path string
            as the one passed in the parameters and returns its column data.
            If no data has been assigned to an entry found so far,
            an empty string is returned.
            If there is no entry with the searched path string,
            an empty string will also be returned.

            This method is called when another document is added to the list.
            If the path to the new document already exists in the list,
            then the same data should be noted for this new entry.

        @param[in] str_path   String with the filename
        @param[in] num_column Column number, containing the data
        @return               Data string
        '''
        str_data = ""
        treeiter = self._lstore_cat_documents.get_iter_first()
        while treeiter is not None:
            if self._lstore_cat_documents.get_value(treeiter, self._COLUMN_PATH) == str_path:
                str_data = self._lstore_cat_documents.get_value(treeiter, num_column)
                break
            treeiter = self._lstore_cat_documents.iter_next(treeiter)
        return str_data


    def _scan_cat_list_for_diff_active_entries(self):
        '''
        @brief Scan the concatenate list store for active entries with different filenames.
        @details
            This information is required to update the Document ID combo box.

        @see _update_cat_combobox_document_id()

        @retval 0 No active entry was found or list is empty
        @retval 1 Active entries with one or the same document path was found
        @retval 2 Active entries with two (or more) various document path was found
        '''
        treeiter_A = self._lstore_cat_documents.get_iter_first()
        while treeiter_A is not None:
            if self._lstore_cat_documents.get_value(treeiter_A, self._COLUMN_ACTIVE) == True:
                str_path_A = self._lstore_cat_documents.get_value(treeiter_A, self._COLUMN_PATH)
                treeiter_B = treeiter_A
                while treeiter_B is not None:
                    if self._lstore_cat_documents.get_value(treeiter_B, self._COLUMN_ACTIVE) == True:
                        str_path_B = self._lstore_cat_documents.get_value(treeiter_B, self._COLUMN_PATH)
                        if str_path_A != str_path_B:
                            return 2  # a second active document path was found
                    treeiter_B = self._lstore_cat_documents.iter_next(treeiter_B)
                return 1              # no second active document path was found, but a first was
            treeiter_A = self._lstore_cat_documents.iter_next(treeiter_A)
        return 0                      # no active document was found


    def _update_cat_combobox_document_id(self):
        '''
        @brief Update the Concatenate document ID combo box.
        @details
            This method is always called when something in the list has changed.

        @see _scan_cat_list_for_diff_active_entries()
        '''
        active_documents = self._scan_cat_list_for_diff_active_entries()
        if active_documents == 0:
            self._cbox_cat_document_id.set_active_id(None)
            self._cbox_cat_document_id.set_sensitive(False)
            self._entry_cat_document_id.set_text("")
            self._entry_cat_document_id.set_placeholder_text("")
        elif active_documents == 1:
            self._cbox_cat_document_id.set_active_id('new')
            self._cbox_cat_document_id.set_sensitive(False)
            self._entry_cat_document_id.set_text("")
            self._entry_cat_document_id.set_placeholder_text("Keep document ID")
        else:
            self._cbox_cat_document_id.set_active_id('first') # TODO: restore old ID
            self._cbox_cat_document_id.set_sensitive(True)


    # Handler for all source document pages
    def on_fcbutton_file_set(self, fcbutton):
        '''
        @brief Signal Handler: On FileChooserButton File set
        @details
            Sets the tool tip to the actual selected path.

        @param[in] fcbutton The object which received the signal
        '''
        fcbutton.set_tooltip_markup(self._STR_PATH + ": '" + fcbutton.get_filename() + "'" )


    # Handler for sub page Burst
    def on_section_burst_changed(self, *args):
        '''
        @brief   Signal Handler: On section Burst has any widget changed.
        @details
            This signal handler is called from any (useful) editable widget
            of the Burst section when it has changed its state
            and when section Burst gets visible.
            The handler calls a function to get the PDFtk burst template string,
            assembled from the states of the widgets in section Burst
            and makes this string visible.

        @see self._refresh_burst_template()
        '''
        num_pages = self._source_document_pages

        if num_pages == 0:
            self._rbutton_burst_manual.set_active(True)
        #else:
        #    self._rbutton_burst_auto.set_active(True)

        self._refresh_burst_template(num_pages)


    def on_rbutton_burst_manual_toggled(self, button):
        '''
        '''
        if button.get_active() == True:
            self._sbutton_burst_manual.set_sensitive(True)
        else:
            self._sbutton_burst_manual.set_sensitive(False)


    def on_entry_burst_prefix_icon_pressed(self, entry, icon_pos, event):
        '''
        '''
        if icon_pos == Gtk.EntryIconPosition.SECONDARY:
            entry.set_text("")


    def on_entry_burst_suffix_icon_pressed(self, entry, icon_pos, event):
        '''
        '''
        if icon_pos == Gtk.EntryIconPosition.SECONDARY:
            entry.set_text("")


    def _refresh_burst_template(self, num_pages):
        '''
        Creates the burst pattern for the PDFtk command.
        '''
        str_prefix    = self._entry_burst_prefix.get_text()
        str_suffix    = self._entry_burst_suffix.get_text()
        str_digits    = "3" # set three digits a fall back
        cmd_base      = "d" # set decimal as fall back base
        num_base      = 10  # set decimal as fall back base
        str_extension = ""

        if not str_prefix:
            str_prefix = self._entry_burst_prefix.get_placeholder_text()

        #TODO: Gtk.ComboBox.use get_active_id()
        treeiter = self._cbox_burst_base.get_active_iter()
        if treeiter is not None:
            treemodel = self._cbox_burst_base.get_model()
            str_id = treemodel[treeiter][self._COLUMN_ID]
            if str_id == 'oct':
                num_base = 8
                cmd_base = pdftk.COUNT_OCT
            elif str_id == 'hex':
                num_base = 16
                cmd_base = pdftk.COUNT_HEX
            else:
                num_base = 10
                cmd_base = pdftk.COUNT_DEC

        if self._rbutton_burst_auto.get_active() == True:
            if num_pages > 0:
                str_digits = str(strings.calculate_digits_of_number_for_base(num_base, num_pages))
            else:
                str_digits = "3" # fallback value
        else:
            str_digits = str(int(self._sbutton_burst_manual.get_value()))

        if self._switch_burst_ext.get_active() == True:
            str_extension = '.pdf'

        str_template = str_prefix + pdftk.TEMPLATE + str_digits + cmd_base \
                + str_suffix + str_extension

        self._label_burst_template.set_label(str_template)


    # Signal Handler: Sub Page Attachment
    def on_toolbutton_attach_add_file_clicked(self, tool_button):
        '''
        Gtk.TreeModel() is the base class of Gtk.ListStore()

        # first get selection of the tree view
        treeselection = self._tview_attach_files.get_selection()

        # then get selected rows of the list store
        treemodel, treepaths = treeselection.get_selected_rows()
        '''
        response = self._fcdialog_open_all_multi.run()
        self._fcdialog_open_all_multi.hide()

        if response == Gtk.ResponseType.OK:
            str_filenames = self._fcdialog_open_all_multi.get_filenames()

            treeselection = self._tview_attach_files.get_selection()
            treemodel, treepaths = treeselection.get_selected_rows()

            if treepaths == []:
                treeiter_selected = None
            else:
                treeiter_selected = treemodel.get_iter(treepaths[0]) # get the first iterator

            for str_filename in str_filenames:
                str_file_basename = GLib.path_get_basename(str_filename)
                str_tooltip = self._STR_PATH + ": '" + str_filename + "'"

                if treeiter_selected == None:
                    treeiter_new = self._lstore_attach_files.append(None)
                else:
                    treeiter_new = self._lstore_attach_files.insert_before(
                            treeiter_selected)

                self._lstore_attach_files[treeiter_new] = [
                        True,
                        str_filename,
                        str_file_basename,
                        str_tooltip]


    def on_toolbutton_attach_remove_file_clicked(self, tool_button):
        '''
        Generate a list of iterators form the tree paths first.
        Delete then the list store entries by using this iterator list.
        '''
        treeselection = self._tview_attach_files.get_selection()
        tree_model_selected, tree_paths_selected = treeselection.get_selected_rows()

        treeiters_selected = []
        for tree_path_selected in tree_paths_selected:
            treeiters_selected.append(tree_model_selected.get_iter(tree_path_selected))
        for treeiter_selected in treeiters_selected:
            self._lstore_attach_files.remove(treeiter_selected)


    def on_rtoolbutton_attach_page_toggled(self, button):
        '''
        '''
        if button.get_active() == True:
            self._toolitem_attach_page.set_sensitive(True)
        else:
            self._toolitem_attach_page.set_sensitive(False)


    def on_crenderer_attach_add_toggled(self, cell_renderer_toggle, str_tview_path):
        '''
        @brief Signal Handler for a CellRenderer Toggle in section Attach: The Active button was toggled.
        @details
            Toggles the boolean value in the concatenate list store.

        @param[in] crtoggle       The object which received the signal _(not used)_
        @param[in] str_tview_path String representation of Gtk.TreePath describing the event location
        '''
        treeiter = self._lstore_attach_files.get_iter(str_tview_path)
        value = self._lstore_attach_files.get_value(treeiter, self._COLUMN_ADD)
        self._lstore_attach_files.set_value(treeiter, self._COLUMN_ADD, not value)


    # Handler for sub page Tools
    def on_rbutton_tools_ddata_toggled(self, button):
        '''
        @brief   Tools; Signal Handler: on RadioButton "Dump Data" toggled
        @details
            Makes associated widgets sensitive.
        '''
        if button.get_active() == True:
            self._cbutton_tools_ddata_utf8.set_sensitive(True)
        else:
            self._cbutton_tools_ddata_utf8.set_sensitive(False)


    def on_rbutton_tools_ddfields_toggled(self, button):
        '''
        @brief   Tools; Signal Handler: on RadioButton "Dump Data Fields" toggled
        @details
            Makes associated widgets sensitive.
        '''
        if button.get_active() == True:
            self._cbutton_tools_ddfields_utf8.set_sensitive(True)
        else:
            self._cbutton_tools_ddfields_utf8.set_sensitive(False)


    def on_rbutton_tools_update_info_toggled(self, button):
        '''
        @brief   Tools; Signal Handler: on RadioButton "Update Info" toggled
        @details
            Makes associated widgets sensitive.
        '''
        if button.get_active() == True:
            self._cbutton_tools_update_info_utf8.set_sensitive(True)
            self._box_tools_update_info.set_sensitive(True)
        else:
            self._cbutton_tools_update_info_utf8.set_sensitive(False)
            self._box_tools_update_info.set_sensitive(False)


    def on_rbutton_tools_fform_toggled(self, rbutton):
        '''
        @brief   Tools; Signal Handler: on RadioButton "FillForm" toggled
        @details
            Makes associated widgets sensitive.
        '''
        if rbutton.get_active() == True:
            self._box_tools_fform_file.set_sensitive(True)
            self._box_tools_fform_options.set_sensitive(True)
        else:
            self._box_tools_fform_file.set_sensitive(False)
            self._box_tools_fform_options.set_sensitive(False)


    # Signal Handler for popover menu Restrictions
    def on_entry_rest_passwd_icon_pressed(self, entry, icon_pos, event):
        '''
        @brief Signal Handler for icon press events of restriction Password entries.
        @details
            A press on the primary entry icon, toggles the visibility of the text.
            One mode shows the password stared, the other in readable text.
            A press on the secondary entry icon, resets (deletes) the password.
        @par
            This signal hander is called by:
            - `entry_rest_passwd_user`
            - `entry_rest_passwd_owner`

        @param[in] entry    The object which received the signal
        @param[in] icon_pos The position of the clicked icon
        @param[in] event    The button press event _(not used)_
        '''
        if icon_pos == Gtk.EntryIconPosition.PRIMARY:
            visibility = entry.get_visibility()
            entry.set_visibility(not visibility)
            if entry.get_visibility() == True:
                entry.set_icon_from_icon_name(Gtk.EntryIconPosition.PRIMARY, "view-reveal-symbolic")
            else:
                entry.set_icon_from_icon_name(Gtk.EntryIconPosition.PRIMARY, "view-conceal-symbolic")
        elif icon_pos == Gtk.EntryIconPosition.SECONDARY:
            entry.set_text("")


    def on_rbutton_rest_enc_toggled(self, rbutton):
        '''
        @brief Signal Handler for toggle events of the encryption radio buttons.
        @details
            As soon as one of the two encryption algorithms is selected,
            the frame Permissions is made available (sensitive).
        @par
            This signal handler is called by:
            - `rbutton_rest_enc_none`
            - `rbutton_rest_enc_rc4_40bit`
            - `rbutton_rest_enc_rc4_128bit`

        @param[in] rbutton The object which received the signal _(not used)_
        '''
        if self._rbutton_rest_enc_none.get_active() == True:
            self._frame_rest_perm.set_sensitive(False)
        else:
            self._frame_rest_perm.set_sensitive(True)


    def on_cbutton_rest_perm_copy_contents_toggled(self, cbutton):
        '''
        @brief Signal Handler for toggle event of check button Copy Content.
        @details
            If Copy Content is allowed, Screen readers are also allowed.
            It is not possible to deselect screen readers
            if copying of content is allowed.
            Therefore, the Screen reader check box is also active
            as soon as Copy Content becomes active.
            It is then not possible to deactivate screen readers.
            Only if Copy Content is deactivated
            is it possible to deactivate screen readers.

        @see on_cbutton_rest_perm_modify_contents_toggled()
        @see on_cbutton_rest_perm_modify_annots_toggled()

        @param[in] cbutton The object which received the signal
        '''
        if cbutton.get_active() == True:
            self._cbutton_rest_perm_screen_readers.set_sensitive(False)
            self._cbutton_rest_perm_screen_readers.set_active(True)
        else:
            self._cbutton_rest_perm_screen_readers.set_sensitive(True)


    def on_cbutton_rest_perm_modify_contents_toggled(self, cbutton):
        '''
        @brief Signal Handler for toggle event of check button Modify Contents.
        @details
            If Modify Contents is allowed, Assembly Contents is also allowed.
            It is not possible to deny Assembly, when Modify Content is allowed.

        @see on_cbutton_rest_perm_copy_contents_toggled()
        @see on_cbutton_rest_perm_modify_annots_toggled()

        @param[in] cbutton The object which received the signal
        '''
        if cbutton.get_active() == True:
            self._cbutton_rest_perm_assembly_contents.set_sensitive(False)
            self._cbutton_rest_perm_assembly_contents.set_active(True)
        else:
            self._cbutton_rest_perm_assembly_contents.set_sensitive(True)


    def on_cbutton_rest_perm_modify_annots_toggled(self, cbutton):
        '''
        @brief Signal Handler for toggle event of check button Modify Annotations.
        @details
            If Modify Annotations is allowed, Fill In Annotations is also allowed.
            It is not possible to deny Fill In, when Modify Annotations is allowed.

        @see on_cbutton_rest_perm_copy_contents_toggled()
        @see on_cbutton_rest_perm_modify_contents_toggled()

        @param[in] cbutton The object which received the signal
        '''
        if cbutton.get_active() == True:
            self._cbutton_rest_perm_fill_in_annots.set_sensitive(False)
            self._cbutton_rest_perm_fill_in_annots.set_active(True)
        else:
            self._cbutton_rest_perm_fill_in_annots.set_sensitive(True)


    # Handler used by widgets from various sections
    def on_entry_rest_passwd_changed(self, editable):
        '''
        @brief Signal Handler: On entry Restrictions Password changed
        @details
            This handler is called for every single character
            entered in one the restriction password entries.
            It checks if the entered password does not contain any characters
            that do not correspond to the Latin-1 encoding.
            If it does, the entry is colored red (style class: "error").

        @param[in] editable The object which received the signal
        '''
        str_passwd = editable.get_text()
        if self._str_is_latin1_encoded(str_passwd) == True:
            editable.get_style_context().remove_class("error")
        else:
            editable.get_style_context().add_class("error")
            cli.debug_message("WARNING: illegal character in restriction password")



# Copyright (C) Martin Singer <martin.singer@web.de>
#
# This file is part of PDF Chain.
#
# PDF Chain is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PDF Chain is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PDF Chain.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Martin Singer

"""
@brief PDFtk Command Strings.
@details
    Contains the PDF Toolkit command and options as constant strings.
@todo
    Implement the constant PDFTK as getter function
    and get the value from the preferences

@see <https://www.pdflabs.com/docs/pdftk-man-page/>
@see <https://www.pdflabs.com/docs/pdftk-cli-examples/>
"""


class PDFtk():
    """
    """

    __PDFTK = "/usr/bin/pdftk"

    # @property
    # def PDFTK():
    #     return PDFtk.__PDFTK

    @staticmethod
    def set_command(path):
        """
        """
        PDFtk.__PDFTK = path


    @staticmethod
    def get_command():
        """
        """
        return PDFtk.__PDFTK


    # General
    # PDFTK                 = "pdftk"  # Do not use this! Use the command from the preferences!
    OUTPUT                = "output"
    PASSWORD_INPUT        = "input_pw"

    # Cat and Shuffle
    CAT                   = "cat"
    SHUFFLE               = "shuffle"

    ID_KEEP               = ""
    ID_NEW                = ""
    ID_FIRST              = "keep_first_id" # ID of first PDF
    ID_FINAL              = "keep_final_id" # ID of final PDF

    ROTATION_NORTH        = "north"         #   0 deg
    ROTATION_EAST         = "east"          #  90 deg
    ROTATION_SOUTH        = "south"         # 180 deg
    ROTATION_WEST         = "west"          # 270 deg

    PAGES_ALL             = ""
    PAGES_EVEN            = "even"
    PAGES_ODD             = "odd"

    # Burst
    BURST                 = "burst"
    TEMPLATE              = "%0"            # Complete suffix example: "%04d.pdf"
    EXTENSION             = ".pdf"
    COUNT_OCT             = "o"
    COUNT_DEC             = "d"
    COUNT_HEX             = "x"

    # Watermark / Stamp
    BACKGROUND            = "background"
    BACKGROUND_MULTI      = "multibackground"
    STAMP                 = "stamp"
    STAMP_MULTI           = "multistamp"

    # Attachment
    ATTACH_FILES          = "attach_files"
    TO_PAGE               = "to_page"

    # Permissions
    ALLOW                 = "allow"
    FEATURES_ALL          = "AllFeatures"

    PRINTING              = "Printing"
    DEGRADED_PRINTING     = "DegradedPrinting"
    MODIFY_CONTENTS       = "ModifyContents"
    ASSEMBLY              = "Assembly"
    COPY_CONTENTS         = "CopyContents"
    SCREEN_READERS        = "ScreenReaders"
    MODIFY_ANNOTATIONS    = "ModifyAnnotations"
    FILL_IN               = "FillIn"

    PASSWORD_OWNER        = "owner_pw"
    PASSWORD_USER         = "user_pw"

    #ENCRYPT_NONE          = ""
    ENCRYPT_40BIT         = "encrypt_40bit"
    ENCRYPT_128BIT        = "encrypt_128bit"

    # Tools
    UNPACK_FILES          = "unpack_files"
    COMPRESS              = "compress"
    UNCOMPRESS            = "uncompress"
    DUMP_DATA_ANNOTS      = "dump_data_annots"
    DUMP_DATA_FIELDS      = "dump_data_fields"
    DUMP_DATA_FIELDS_UTF8 = "dump_data_fields_utf8"
    DUMP_DATA             = "dump_data"
    DUMP_DATA_UTF8        = "dump_data_utf8"
    UPDATE_INFO           = "update_info"
    UPDATE_INFO_UTF8      = "update_info_utf8"
    GENERATE_FDF          = "generate_fdf"
    FILL_FORM             = "fill_form"
    FLATTEN               = "flatten"
    NEED_APPEARANCES      = "need_appearances"
    DROP_XFA              = "drop_xfa"



# Copyright (C) Martin Singer <martin.singer@web.de>
#
# This file is part of PDF Chain.
#
# PDF Chain is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PDF Chain is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PDF Chain.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Martin Singer

"""
@brief The Preferences Module.
@details
    The application manages the various windows and dialogs.
    It contains the preferences and about dialog
    as well as multiple application window instances.
"""

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

import configparser

import pdfchain.paths as paths
from pdfchain.cmd_pdftk import PDFtk as pdftk


class Preferences():
    """
    @brief Preferences Class.
    @details
        Loads, stores and saves the user preference values
        as well as running the preferences GUI dialog.
    """

    _PATH_PREF_FILE = paths.get_config_file()
    _PATH_PDFTK_DEFAULT_CMD = "/usr/bin/pdftk"

    # Gtk widgets
    _dialog = None
    _entry_path_pdftk = None

    # ConfigParser object
    _pref = None


    def __init__(self):
        """
        @brief Constructing a Preferences Object.
        @details
            Loads the UI file and builds the dialog widgets.
            Creates the configuration parser object, loads and exports the preferences.
        """
        builder = Gtk.Builder()
        builder.add_from_file(paths.get_ui_file("dialogs.ui"))
        self._entry_path_pdftk = builder.get_object("entry_path_pdftk")
        self._entry_path_pdftk.set_placeholder_text(self._PATH_PDFTK_DEFAULT_CMD)
        self._dialog = builder.get_object("dialog_preferences")

        if self._pref is None:
            self._pref = configparser.ConfigParser()
            self._load_file()
            self._export_values()


    def _save_file(self):
        """
        @brief Get preference values from dialog widgets and save preferences into file.
        @details
            Sets the values in the configuration parser object
            and saves the preferences file, using this object.
        """
        path_pdftk = self._entry_path_pdftk.get_text()
        self._pref['PATH'] = {
            'pdftk': path_pdftk
            }

        with open(self._PATH_PREF_FILE, 'w') as preffile:
            self._pref.write(preffile)


    def _load_file(self):
        """
        @brief Load preference values from file or set default values.
        @details
            If the preferences file was found, the configuration parser object
            reads the preferences file.
            If the preferences file was not found, default variables are set.
            The defined variables are set to the dialog widgets.
        """
        self._pref.read(self._PATH_PREF_FILE)
        path_pdftk = self._pref['PATH']['pdftk']
        if not path_pdftk:
            path_pdftk = self._PATH_PDFTK_DEFAULT_CMD
            self._pref['PATH'] = {
                'pdftk': path_pdftk
                }
        self._entry_path_pdftk.set_text(path_pdftk)


    def _export_values(self):
        """
        @brief Export the preferences variables to the corresponding modules.
        @details
            External module methods and functions are called to transfer
            preferences, stored in the configuration parser object.
        """
        path_pdftk = self._pref['PATH']['pdftk']
        pdftk.set_command(path_pdftk)


    def run_dialog(self, parent):
        """
        @brief Run the preferences dialog.
        @details
            This method calls the methods which are required
            to load, show, save and export the preference values.

        @param[in] parent The parent window or None
        """
        self._load_file()
        self._dialog.set_transient_for(parent)
        self._dialog.run()
        self._save_file()
        self._export_values()
        self._dialog.hide()


    # def get_path(self, str_command):
    #         return self._pref['PATH'][str_command]



# Copyright (C) Martin Singer <martin.singer@web.de>
#
# This file is part of PDF Chain.
#
# PDF Chain is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PDF Chain is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PDF Chain.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Martin Singer


'''
@brief PDF Chain string module.
@details
    This module contains classed with string constants
    or functions to process strings.
'''

import re

# import cli
import pdfchain.cli


def validate_page_selection_string(str_input):
    '''
    @brief   Validate page selection string from Concatenate File Filter entry.
    @details
        This functions normalizes and validates page selection strings.
        Before the validation the string becomes normalized.
        Normalization means the string becomes transformed
        to a unique PDFtk convented pattern.
        This is possible for PDFtk standardized strings
        and when the string equates a clear, comma separated pattern style.
        Mixing the styles is possible.

    @par Strings and Characters
        - `end`: indicates last page of document
        - `r`:   indicates reverse
        - `-`:   indicates connected pages
        - `,`:   indicates a separation of logical entities
        - ` `:   indicates a separation of logical

    @par Examples
        | Input String             | Output String   | Valid | Annotation                                            |
        | :----------------------- | :-------------- | :---- | :---------------------------------------------------- |
        | `1-end 2 1 r9`           | `1-end 2 1 r9`  | True  | clean PDFtk style                                     |
        | `1 - end, 2, 1, r9`      | `1-end 2 1 r9`  | True  | clean comma based style                               |
        | `end- 1 3 , r7 -31`      | `end-1 3 r7-31` | True  | tolerant to mixed separator styles                    |
        | `end-1, , 3  , r7  - 31` | `end-1 3 r7-31` | True  | tolerant to multiple comma and multiple spaces        |
        | `1-e n d, 4 - r 3`       | ``              | False | invalid spaces in indicator strings                   |
        | `38r`                    | ``              | False | indicator 'r' behind number is invalid                |
        | `1-end, 4, - 2`          | ``              | False | invalid comma separated sub entities _(is not `4-2`)_ |
        | `1 - en, a, x2`          | ``              | False | unknown strings and unknown characters                |
        | `1 - 4-1`                | ``              | False | double connections                                    |
        | `r r4, end end, r end`   | ``              | False | double indicator                                      |
        | `3r4, 9end, rend`        | ``              | False | missing separator or connector characters             |

    @par RegEx for PDFtk normalized string
        - from (required):
            - `^[1-9][0-9]*`
            - `^r[1-9][0-9]*`
            - `^end`
        - to (optional):
            - `-[1-9][0-9]*`
            - `-r[1-9][0-9]*`
            - `-end`

    @param[in]       The user file selection string
    @retval[0] True  When selection string is valid
    @retval[0] False When selection string is not valid
    @return[1]       Returns a PDFtk page selection command string, when file selection is interpretable and correct
    @retval[1] ""    Returns the input string, if the input string is empty or not interpretable
    '''
    REGEX = r"^((r?[1-9][0-9]*)|end)(-((r?[1-9][0-9]*)|end))?"
    valid = True
    str_output = ""

    # If string is empty, set default string
    if not str_input:
        str_base = "1-end"  #TODO: define this string as a constant
    else:
        str_base = str_input.lower()

    # Normalize string (create PDFtk command string)
    # (do not remove spaces as separators - just multiple spaces)
    arr_parts = str_base.split(',')
    for str_part in arr_parts:
        str_part = str_part.strip(',')             # strip leading and trailing commas from string
        str_part = str_part.strip(' ')             # strip leading and trailing spaces from string
        str_part = re.sub(r"\s+", r" ", str_part)  # remove multiple spaces
        str_part = str_part.replace(' -', '-')     # remove leading  spaces from hyphen
        str_part = str_part.replace('- ', '-')     # remove trailing spaces from hyphen
        str_output += ' ' + str_part

    # Validate normalized string
    arr_parts = str_output.split(None)
    for str_part in arr_parts:
        if re.fullmatch(REGEX, str_part) is None:
            valid = False
            #cli.debug_message("ERROR: Expression '{}'"
            #        " is no valid page selection pattern".format(str_part)) #DEBUG

    if valid == True:
        return valid, str_output
    else:
        return valid, str_input


def create_handle_from_number(number):
    '''
    @brief   Create a handle from an integer number.
    @details
        A handle is a letter or a letter string, representing a document path.
        It is used to assign various options to a document
        in the concatenate command.
        In this function a decimal number (base 10) becomes converted
        to a letter number (base 26), returned as string.
        - A 'A' is equivalent to a '0'
        - A 'Z' is equivalent to a '25'
        Leading 'A's are not visible as well leading '0's are not.

    @param[in] number The document number (id)
    @return           A string with a handle
    '''
    BASE = 26
    handle = ""

    # Convert number from base 10 [0 to 9] to base 26 [A to Z]
    while number >= BASE:
        handle += chr(ord('A') + number % BASE)
        number //= BASE

    handle += chr(ord('A') + number % BASE)
    return handle


def count_pages_of_document(path):
    '''
    @brief   Count the amount of pages of a PDF document.
    @details
        The document specified in the transferred path is opened
        and a memory buffer is read in binary. In this buffer,
        the respective number of four significant patterns is counted
        and calculated. This result is the page number of the document.
    @see <https://docs.python.org/3/library/stdtypes.html#bytes-and-bytearray-operations>

    @param[in] path The path to the document file
    @retval    0    The number of pages cannot be determined
    @return         The number of pages determined
    '''
    TYPE_PAGE_A = b'/Type/Page'
    TYPE_PAGE_B = b'/Type /Page'
    TYPE_PAGE_C = b'/Type/Pages'
    TYPE_PAGE_D = b'/Type /Pages'

    num_pages = 0

    try:
        with open(path, mode='rb') as source_file:
            file_content = source_file.read()
            readable = True

            count_A = file_content.count(TYPE_PAGE_A)
            count_B = file_content.count(TYPE_PAGE_B)
            count_C = file_content.count(TYPE_PAGE_C)
            count_D = file_content.count(TYPE_PAGE_D)

            num_pages = count_A + count_B - count_C - count_D
    except IOError as io_error:
        if io_error.errno == errno.ENOENT:
            cli.msg("ERROR: file '{}' does not exist!".format(path))
        elif io_error.errno == errno.EACCES:
            cli.msg("ERROR: file '{}' cannot be read!".format(path))
        else:
            cli.msg("ERROR: Unknown error in function 'count_pages_of_document()'")
            #TODO: maybe rise exception

    return num_pages


def calculate_digits_of_number_for_base(base, number):
    '''
    @brief   Calculates the number of digits of a decimal number to another base.
    @details
        This function is needed in the burst section, when the number of digits
        for the different bases must be calculated automatically.
        It calculates the number of digits that this number needs for another base.
    @warning
        Does not work correct for negative numbers and bases!
    @note
        This function uses the `__floordiv__` operator `//`,
        for a "classic division" with truncated decimal places.
        `number //= base` equals `number = int(number / base)`

    @param[in] base   The base to which the number of digits is to be calculated
    @param[in] number The decimal number
    @retval    0      The number of digits cannot be determined
    @return           The number of digits determined
    '''
    digits = 0

    while number > 0:
        number //= base
        digits += 1

    return digits


def quote_and_escape_string(string):
    '''
    @brief Quote and escape strings for the command line.
    @details
        This function is used to prepare strings like filenames
        and passwords for the command line.
    @note About quoting
        Inside a single quoted string is the single quote character `'`
        the only character which has to be handled.
        But there is no escape character which works inside a single quoted string.
        Therefore a single quoted string must be ended, the single quote escaped
        and the rest of the string restarts quoted again.

    @param[in] string The string to quote
    return            The quoted and escaped string
    '''
    return "'" + string.replace(r"'", r"'\''") + "'"



# Copyright (C) Martin Singer <martin.singer@web.de>
#
# This file is part of PDF Chain.
#
# PDF Chain is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PDF Chain is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PDF Chain.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Martin Singer

'''
 @brief   File Chooser Dialog File Filter
 @details
    The file filters do not work very well in Glade UI definitions.
    So I implemented them in code.

 @note Inadequacies with the file filters using the UI definitions
     - Multiple file filters are not possible
     - File filter can not be named
     - File filter can not be shared
       (have to be extra defined file filter for every single dialog)
'''

import os # for `os.path.splitext()`
import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk


# File Filter
class FileFilter_All(Gtk.FileFilter):
    """
    @brief File Filter: All files
    """

    def __init__(self):
        super().__init__()
        self.set_name("All Files")
        self.add_pattern("*")


class FileFilter_PDF(Gtk.FileFilter):
    """
    @brief File Filter: Portable Document Format file
    """

    def __init__(self):
        super().__init__()
        self.set_name("PDF Files")
        self.add_mime_type("application/pdf")
        self.add_pattern("*.pdf")


class FileFilter_Text(Gtk.FileFilter):
    """
    @brief File Filter: Text files
    """

    def __init__(self):
        super().__init__()
        self.set_name("Text Files")
        self.add_mime_type("text/plain")
        self.add_pattern("*.txt")


class FileFilter_FDF(Gtk.FileFilter):
    """
    @brief File Filter: Forms Data Format files
    @details
        - Pattern `*.fdf`:  Forms Data Format file
        - Pattern `*.xfdf`: XML Forms Data Format file
    @note
        PDFtk can create form data exclusively in FDF format.
    """

    def __init__(self):
        super().__init__()
        self.set_name("FDF Files")
        self.add_pattern("*.fdf")
        self.add_pattern("*.xfdf")


class FileFilter_Dump(Gtk.FileFilter):
    """
    @brief File Filter: Metadata Dump files
    """

    def __init__(self):
        super().__init__()
        self.set_name("Metadata Dump Files")
        self.add_pattern("*.dump")


# FileChooserDialogs
class FCDialog_SaveDoc(Gtk.FileChooserDialog):
    """
    @brief File Chooser Dialog: Save PDF Files
    """

    def __init__(self, transient):
        super().__init__("Save Document", transient, Gtk.FileChooserAction.SAVE,
                ("_Save", Gtk.ResponseType.OK, "_Cancel", Gtk.ResponseType.CANCEL))
        self.set_create_folders(True)
        self.set_current_name("Untitled.pdf")
        self.set_do_overwrite_confirmation(True)
        self.add_filter(FileFilter_PDF())
        self.add_filter(FileFilter_All())


class FCDialog_OpenDoc(Gtk.FileChooserDialog):
    """
    @brief File Chooser Dialog: Open PDF Files
    """

    def __init__(self, transient):
        super().__init__("Open Document", transient, Gtk.FileChooserAction.OPEN,
                ("_Open", Gtk.ResponseType.OK, "_Cancel", Gtk.ResponseType.CANCEL))
        self.add_filter(FileFilter_PDF())
        self.add_filter(FileFilter_All())


class FCDialog_OpenDocMulti(Gtk.FileChooserDialog):
    """
    @brief File Chooser Dialog: Open multiple PDF Files
    """

    def __init__(self, transient):
        super().__init__("Open Documents", transient, Gtk.FileChooserAction.OPEN,
                ("_Open", Gtk.ResponseType.OK, "_Cancel", Gtk.ResponseType.CANCEL))
        self.set_select_multiple(True)
        self.add_filter(FileFilter_PDF())
        self.add_filter(FileFilter_All())


class FCDialog_SaveFDF(Gtk.FileChooserDialog):
    """
    @brief File Chooser Dialog: Save Forms Data Format files
    @note
        PDFtk can create form data exclusively in FDF format.
        Therefore no file filter for the XFDF format is added
        to the save FDF file dialog.
    """

    def __init__(self, transient):
        super().__init__("Save FDF file", transient, Gtk.FileChooserAction.SAVE,
                ("_Save", Gtk.ResponseType.OK, "_Cancel", Gtk.ResponseType.CANCEL))
        self.set_create_folders(True)
        self.set_current_name("Untitled.fdf")
        self.set_do_overwrite_confirmation(True)
        self.add_filter(FileFilter_FDF())
        self.add_filter(FileFilter_Text())
        self.add_filter(FileFilter_All())


#class FCDialog_OpenFDF(Gtk.FileChooserDialog):
#    """
#    @brief File Chooser Dialog: Open FDF
#    """
#
#    def __init__(self, transient):
#        super().__init__("Open FDF file", transient, Gtk.FileChooserAction.OPEN,
#                ("_Open", Gtk.ResponseType.OK, "_Cancel", Gtk.ResponseType.CANCEL))
#        self.add_filter(FileFilter_FDF())
#        self.add_filter(FileFilter_XFDF())
#        self.add_filter(FileFilter_Text())
#        self.add_filter(FileFilter_All())


class FCDialog_SaveDump(Gtk.FileChooserDialog):
    """
    @brief File Chooser Dialog: Save Metadata files
    """

    def __init__(self, transient):
        super().__init__("Save Metadata Dump File", transient, Gtk.FileChooserAction.SAVE,
                ("_Save", Gtk.ResponseType.OK, "_Cancel", Gtk.ResponseType.CANCEL))
        self.set_create_folders(True)
        self.set_current_name("Untitled.dump")
        self.set_do_overwrite_confirmation(True)
        self.add_filter(FileFilter_Dump())
        self.add_filter(FileFilter_Text())
        self.add_filter(FileFilter_All())


#class FCDialog_OpenDump(Gtk.FileChooserDialog):
#    """
#    @brief File Chooser Dialog: Open Metadata files
#    """
#
#    def __init__(self, transient):
#        super().__init__("Open Metadata Dump File", transient, Gtk.FileChooserAction.OPEN,
#                ("_Open", Gtk.ResponseType.OK, "_Cancel", Gtk.ResponseType.CANCEL))
#        self.add_filter(FileFilter_Dump())
#        self.add_filter(FileFilter_Text())
#        self.add_filter(FileFilter_All())


class FCDialog_OpenAllMulti(Gtk.FileChooserDialog):
    """
    @brief File Chooser Dialog: Open multiple (various) files
    """

    def __init__(self, transient):
        super().__init__("Open Various Files", transient, Gtk.FileChooserAction.OPEN,
                ("_Open", Gtk.ResponseType.OK, "_Cancel", Gtk.ResponseType.CANCEL))
        self.set_select_multiple(True)
        self.add_filter(FileFilter_All())
        self.add_filter(FileFilter_PDF())
        self.add_filter(FileFilter_Text())


class FCDialog_SelectFolder(Gtk.FileChooserDialog):
    """
    @brief File Chooser Dialog: Select Folder
    """

    def __init__(self, transient):
        super().__init__("Select Folder", transient, Gtk.FileChooserAction.SELECT_FOLDER,
                ("_Select", Gtk.ResponseType.OK, "_Cancel", Gtk.ResponseType.CANCEL))
        self.set_do_overwrite_confirmation(True)
        self.set_create_folders(True)


# Functions
def fchooser_run(fcdialog):
    """
    @brief Helper function to run file choosers.
    @details
        This function has the standard procedure to run a file choosers.
        Works for file chooser dialogs and buttons.

    @param[in] fcdialog     The FileChooserDialog object to run()
    @return    str_filename String of chosen filename path
    @retval    ""           Empty string when "Cancel" was clicked
    """
    str_filename = ""
    response = fcdialog.run()
    fcdialog.hide()
    if response == Gtk.ResponseType.OK:
        str_filename = fcdialog.get_filename()
        if str_filename == None:
            cli.debug_message("WARNING: str_filename is None")
            str_filename = ""
    return str_filename



#!/usr/bin/env python3

# Copyright (C) Martin Singer <martin.singer@web.de>
#
# This file is part of PDF Chain.
#
# PDF Chain is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PDF Chain is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PDF Chain.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Martin Singer

"""
 @brief The Main Module.
 @details
    This module is called when starting the application
    by calling the directory or the ZIP package name.
"""

import sys

import pdfchain.app as app


def main():
    """
    @brief Starting the Application
    """
    application = app.Application()
    application.run(sys.argv)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import os
# import sys
import setuptools


def get_data_files():
    data_files = []

    for path, dirs, files in os.walk("share"):
        # target_path = os.path.join("/usr", path)
        # target_path = os.path.join(sys.prefix, path)
        target_path = path
        data_files.append((target_path, [os.path.join(path, f) for f in files]))
    print("data_files: '{}'".format(data_files)) #TEST
    return data_files


def main():
    with open("README.md", "r") as fh:
        long_description = fh.read()

    setuptools.setup(
        name = "pdfchain",
        version = "0.5.0.0 alpha 4",
        author = "Martin Singer",
        author_email = "martin.singer@web.de",
        description = "A graphical user interface for the PDF Toolkit",
        long_description = long_description,
        long_description_content_type = "text/markdown",
        keywords = "PDF",
        url = "https://pdfchain.sourceforge.io/",
        download_url = "https://sourceforge.net/p/pdfchain/neo/ci/master/tree/",
        license='GPLv3+',
        # package_dir = {"": "src"},
        packages = setuptools.find_packages(),
        data_files = get_data_files(),
        # scripts = ["bin/pdfchain"],
        entry_points = {"gui_scripts": ["pdfchain = pdfchain.__main__:main"]},
        classifiers = [
            "Development Status :: 3 - Alpha",
            "Environment :: X11 Applications :: GTK",
            "Intended Audience :: End Users/Desktop",
            "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
            "Operating System :: POSIX",
            "Topic :: Office/Business",
            "Programming Language :: Python :: 3",
        ],
        python_requires = ">=3.6",
        # install_requires = ["PyGObject"],
    )


if __name__ == "__main__":
    main()
